import logging
import re

import pefile
import semver
from packageurl import PackageURL
from pydantic import BaseModel

from labels.model.file import LocationReadCloser
from labels.model.package import Language, Package, PackageType
from labels.model.relationship import Relationship
from labels.model.resolver import Resolver
from labels.parsers.cataloger.generic.parser import Environment

LOGGER = logging.getLogger(__name__)


class DotnetPortableExecutableEntry(BaseModel):
    assembly_version: str | None = None
    legal_copyright: str | None = None
    company_name: str | None = None
    product_name: str | None = None
    product_version: str | None = None
    comments: str | None = None
    internal_name: str | None = None


# Define the regular expressions
space_regex = re.compile(r"[\s]+")
number_regex = re.compile(r"\d")
version_punctuation_regex = re.compile(r"[.,]+")


def parse_version_resource(potable: pefile.PE) -> dict[str, str] | None:
    if not hasattr(potable, "VS_VERSIONINFO"):
        return None

    for idx, _ in enumerate(potable.VS_VERSIONINFO):
        if not hasattr(potable, "FileInfo") or len(potable.FileInfo) <= idx:
            continue

        stringtable_dict = process_file_info(
            potable.FileInfo[idx],  # type: ignore
        )
        if stringtable_dict:
            return stringtable_dict

    return None


def process_file_info(
    file_info: list[pefile.Structure],
) -> dict[str, str] | None:
    for entry in file_info:
        if not hasattr(entry, "StringTable"):
            continue

        stringtable_dict = process_string_table(entry.StringTable)
        if stringtable_dict:
            return stringtable_dict

    return None


def process_string_table(
    string_table: list[pefile.Structure],
) -> dict[str, str]:
    stringtable_dict = {}
    for st_entry in string_table:
        stringtable_dict["LangID"] = st_entry.LangID  # type: ignore
        for key, value in st_entry.entries.items():  # type: ignore
            stringtable_dict[key.decode("utf-8")] = value.decode("utf-8")
    return stringtable_dict


def is_microsoft(version_resources: dict[str, str]) -> bool:
    company_name = version_resources.get("CompanyName", "").lower()
    product_name = version_resources.get("ProductName", "").lower()

    return "microsoft" in company_name or "microsoft" in product_name


def space_normalize(value: str) -> str:
    # Trim leading and trailing whitespace
    value = value.strip()

    if value == "":
        return ""

    # Ensure valid UTF-8 text
    value = value.encode("utf-8", "replace").decode("utf-8")

    value = space_regex.sub(" ", value)

    # Remove other non-space, non-printable characters
    value = re.sub(r"[\x00-\x1f]", "", value)

    # Consolidate all space characters again
    value = space_regex.sub(" ", value)

    # Finally, remove any remaining surrounding whitespace
    return value.strip()


def find_name(version_resources: dict[str, str]) -> str:
    # Define the order of fields to check for the name
    name_fields = [
        "ProductName",
        "FileDescription",
        "InternalName",
        "OriginalFilename",
    ]

    # Check if the version resources are from Microsoft
    if is_microsoft(version_resources):
        # Adjust the order of fields for Microsoft-authored files
        name_fields = [
            "FileDescription",
            "InternalName",
            "OriginalFilename",
            "ProductName",
        ]

    # Iterate over the fields to find a non-empty, normalized name
    for field in name_fields:
        value = space_normalize(version_resources.get(field, ""))
        if value:
            return value

    # Return an empty string if no valid name is found
    return ""


def contains_number(string: str) -> bool:
    return any(char.isdigit() for char in string)


def extract_version(version: str) -> str:
    # Trim leading and trailing whitespace
    version = version.strip()

    out = ""

    # Split the version string into fields and iterate over them
    for index, char in enumerate(version.split()):
        # If the output already has a number but the current segment does not,
        # return the output
        if contains_number(out) and not contains_number(char):
            return out

        # Append the current field to the output
        if index == 0:
            out = char
        else:
            out += " " + char

    return out


def keep_greater_semantic_version(product_version: str, file_version: str) -> str:
    try:
        semantic_product_version = semver.VersionInfo.parse(product_version)
    except ValueError:
        LOGGER.debug(
            ("Unable to create semantic version from portable executable product version %s"),
            product_version,
        )
        return ""

    try:
        semantic_file_version = semver.VersionInfo.parse(file_version)
    except ValueError:
        LOGGER.debug(
            ("Unable to create semantic version from portable executable file version %s"),
            file_version,
        )
        return product_version

    # Make no choice when they are semantically equal
    if semantic_product_version == semantic_file_version:
        return ""

    if semantic_file_version > semantic_product_version:
        return file_version

    return product_version


def punctuation_count(string: str) -> int:
    # Find all matches of the punctuation regex and return their count
    return len(version_punctuation_regex.findall(string))


def find_version(version_resources: dict[str, str]) -> str:
    product_version = extract_version(version_resources.get("ProductVersion", ""))
    file_version = extract_version(version_resources.get("FileVersion", ""))

    semantic_version_compare_result = keep_greater_semantic_version(product_version, file_version)

    if semantic_version_compare_result:
        return semantic_version_compare_result

    product_version_detail = punctuation_count(product_version)
    file_version_detail = punctuation_count(file_version)

    if contains_number(product_version) and product_version_detail >= file_version_detail:
        return product_version

    if contains_number(file_version) and file_version_detail > 0:
        return file_version

    if contains_number(product_version):
        return product_version

    if contains_number(file_version):
        return file_version

    return product_version


def build_dot_net_package(
    version_resource: dict[str, str],
    reader: LocationReadCloser,
) -> Package | None:
    name = find_name(version_resource)
    if not name:
        LOGGER.debug(
            "Unable to find name for portable executable in file %s",
            reader.location.path(),
        )
        return None
    version = find_version(version_resource)
    if not version:
        LOGGER.debug(
            "Unable to find version for portable executable in file %s",
            reader.location.path(),
        )
        return None
    metadata = DotnetPortableExecutableEntry(
        assembly_version=version_resource.get("Assembly Version"),
        legal_copyright=version_resource.get("LegalCopyright"),
        comments=version_resource.get("Comments"),
        internal_name=version_resource.get("InternalName"),
        company_name=version_resource.get("CompanyName"),
        product_name=version_resource.get("ProductName"),
        product_version=version_resource.get("ProductVersion"),
    )

    return Package(
        name=name,
        version=version,
        locations=[reader.location],
        type=PackageType.DotnetPkg,
        language=Language.DOTNET,
        p_url=PackageURL(
            type="nuget",
            namespace="",
            name=name,
            version=version,
            qualifiers={},
            subpath="",
        ).to_string(),
        metadata=metadata,
        licenses=[],
    )


def parse_dotnet_portable_executable(
    _resolver: Resolver | None,
    _env: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    if not reader.location.coordinates:
        return [], []
    try:
        pe_representation = pefile.PE(reader.location.coordinates.real_path, fast_load=False)
    except pefile.PEFormatError:
        return [], []

    version_resource = parse_version_resource(pe_representation)
    if not version_resource:
        LOGGER.debug(
            ("Unable to find version resource for portable executable in file %s"),
            reader.location.path(),
        )
        return [], []
    dotnet_package = build_dot_net_package(version_resource, reader)
    if not dotnet_package:
        LOGGER.debug(
            "Unable to build package for portable executable in file %s",
            reader.location.path(),
        )
        return [], []
    return [dotnet_package], []



import logging
import zipfile
from collections.abc import Callable
from contextlib import suppress
from fnmatch import fnmatch
from zipfile import BadZipFile, ZipFile, ZipInfo

LOGGER = logging.getLogger(__name__)


def normalize_zip_entry_name(entry: str, *, case_insensitive: bool) -> str:
    if case_insensitive:
        entry = entry.lower()
    if not entry.startswith("/"):
        entry = "/" + entry

    return entry


def new_zip_file_manifest(archive_path: str) -> list[ZipInfo]:
    try:
        with ZipFile(archive_path, "r") as myzip:
            return myzip.infolist()
    except BadZipFile:
        return []


def zip_glob_match(
    manifest: list[ZipInfo],
    *,
    case_sensitive: bool,
    patterns: tuple[str, ...],
) -> list[str]:
    result = []

    for pattern in patterns:
        for entry in manifest:
            normalized_entry = normalize_zip_entry_name(
                entry.filename,
                case_insensitive=case_sensitive,
            )
            if entry.filename.endswith(pattern):
                result.append(entry.filename)
            lower_pattern = pattern.lower() if case_sensitive else pattern
            if fnmatch(normalized_entry, lower_pattern):
                result.append(entry.filename)
    return result


def traverse_files_in_zip(
    archive_path: str,
    visitor: Callable[[zipfile.ZipInfo], None],
    *paths: str,
) -> None:
    """Traverse files in a zip file applying a visitor function to each file."""
    with zipfile.ZipFile(archive_path, "r") as zip_reader:
        for path in paths:
            try:
                visitor(zip_reader.getinfo(path))
            except KeyError:
                logging.exception("Unable to find file: %s", path)


def contents_from_zip(archive_path: str, *paths: str) -> dict[str, str]:
    """Extract specified files from a zip archive and return their contents."""
    results: dict[str, str] = {}

    if not paths:
        return results

    def visitor(file: zipfile.ZipInfo) -> None:
        """Visitor function to read the contents of a file in the zip."""
        if file.is_dir():
            logging.error("Unable to extract directories, only files: %s", file.filename)
            return
        with zipfile.ZipFile(archive_path, "r") as zip_reader, zip_reader.open(file) as file_data:
            content = file_data.read()
            with suppress(UnicodeDecodeError):
                results[file.filename] = content.decode(
                    "utf-8",
                )

    traverse_files_in_zip(archive_path, visitor, *paths)
    return results



from decimal import Decimal

from cyclonedx.factory.license import LicenseFactory
from cyclonedx.model import HashType, Property, XsUri
from cyclonedx.model.contact import OrganizationalContact
from cyclonedx.model.license import License
from cyclonedx.model.vulnerability import (
    BomTarget,
    BomTargetVersionRange,
    Vulnerability,
    VulnerabilityAdvisory,
    VulnerabilityRating,
    VulnerabilitySeverity,
)

from labels.model.file import Location
from labels.model.package import HealthMetadata, Package
from labels.output.utils import get_author_info

NAMESPACE = "fluid-attacks"


def add_vulnerabilities(package: Package) -> list[Vulnerability]:
    vulnerabilities = []
    if package.advisories:
        for advisory in package.advisories:
            rating = VulnerabilityRating(
                severity=VulnerabilitySeverity(advisory.severity.lower()),
                score=Decimal(str(advisory.percentile)),
            )

            advisory_obj = (
                [VulnerabilityAdvisory(url=XsUri(url)) for url in advisory.urls]
                if advisory.urls
                else []
            )

            bom_target = BomTarget(
                ref=f"{package.name}@{package.version}",
                versions=[
                    BomTargetVersionRange(
                        version=package.version if not advisory.version_constraint else None,
                        range=advisory.version_constraint,
                    ),
                ],
            )

            additional_properties = [
                Property(name=f"{NAMESPACE}:advisory:epss", value=str(advisory.epss)),
            ]

            vulnerability = Vulnerability(
                bom_ref=f"{package.name}@{advisory.id}",
                id=advisory.id,
                description=advisory.description,
                advisories=advisory_obj,
                ratings=[rating],
                affects=[bom_target],
                properties=additional_properties,
            )

            vulnerabilities.append(vulnerability)

    return vulnerabilities


def add_health_metadata_properties(
    health_metadata: HealthMetadata,
) -> list[Property]:
    properties = []
    if health_metadata.latest_version:
        properties.append(
            Property(
                name=f"{NAMESPACE}:health_metadata:latest_version",
                value=health_metadata.latest_version,
            ),
        )

    if health_metadata.latest_version_created_at:
        properties.append(
            Property(
                name=f"{NAMESPACE}:health_metadata:latest_version_created_at",
                value=str(health_metadata.latest_version_created_at),
            ),
        )

    return properties


def add_locations_properties(locations: list[Location]) -> list[Property]:
    properties = []
    for idx, location in enumerate(locations):
        path = location.path()

        line = (
            location.coordinates.line
            if location.coordinates and location.coordinates.line
            else None
        )
        layer = (
            location.coordinates.file_system_id
            if location.coordinates and location.coordinates.file_system_id
            else None
        )

        properties.append(Property(name=f"{NAMESPACE}:locations:{idx}:path", value=path))

        if line:
            properties.append(Property(name=f"{NAMESPACE}:locations:{idx}:line", value=str(line)))

        if layer:
            properties.append(Property(name=f"{NAMESPACE}:locations:{idx}:layer", value=layer))

    return properties


def add_language_property(package_language: str) -> Property:
    return Property(name=f"{NAMESPACE}:language", value=package_language)


def add_integrity(health_metadata: HealthMetadata) -> list[HashType] | None:
    integrity = health_metadata.artifact.integrity if health_metadata.artifact else None
    if integrity and integrity.algorithm and integrity.value:
        return [HashType.from_hashlib_alg(integrity.algorithm, integrity.value)]
    return None


def add_authors(health_metadata: HealthMetadata) -> list[OrganizationalContact]:
    authors_info = get_author_info(health_metadata)
    return [
        OrganizationalContact(name=name, email=email)
        for name, email in authors_info
        if name or email
    ]


def get_licenses(package_licenses: list[str]) -> list[License]:
    lc_factory = LicenseFactory()
    return [lc_factory.make_from_string(lic) for lic in package_licenses]


def add_component_properties(package: Package) -> list[Property]:
    properties = []

    properties.append(add_language_property(package.language))
    properties.extend(add_locations_properties(package.locations))

    if package.health_metadata:
        properties.extend(add_health_metadata_properties(package.health_metadata))

    return properties



import ast
from collections.abc import Iterator
from contextlib import suppress
from copy import deepcopy

from labels.model.file import Location, LocationReadCloser
from labels.model.package import Package
from labels.model.relationship import Relationship
from labels.model.resolver import Resolver
from labels.parsers.cataloger.cpp.package import new_conan_file_py_dep
from labels.parsers.cataloger.generic.parser import Environment
from labels.parsers.cataloger.graph_parser import Graph, NId, adj, node_to_str, parse_ast_graph


def get_method_requirements(
    *,
    graph: Graph,
    method_id: NId,
    location: Location,
    is_dev: bool,
    requires_attributes: dict[str, str],
) -> Iterator[Package]:
    childs = adj(graph, method_id, depth=-1)
    requirements_nodes = [
        child for child in childs if graph.nodes[child].get("label_type") == "call"
    ]
    for req_node in requirements_nodes:
        node_attrs = graph.nodes[req_node]
        if (c_id := node_attrs.get("label_field_function")) and node_to_str(
            graph,
            c_id,
        ) != requires_attributes["method_requires"]:
            continue
        req_args_id = graph.nodes[req_node]["label_field_arguments"]
        req_args = adj(graph, req_args_id)
        yield new_conan_file_py_dep(
            graph=graph,
            node_id=req_args[1],
            location=location,
            is_dev=is_dev,
        )


def get_attr_requirements(
    *,
    graph: Graph,
    attr_info: dict,
    location: Location,
    is_dev: bool,
) -> Iterator[Package]:
    with suppress(SyntaxError):
        val_id = attr_info["label_field_right"]
        requires_info = graph.nodes[val_id]

        if requires_info["label_type"] == "string":
            requires_eval = ast.literal_eval(node_to_str(graph, val_id))
            yield new_conan_file_py_dep(
                graph=graph,
                node_id=val_id,
                dep_info=requires_eval,
                location=location,
                is_dev=is_dev,
            )
        elif requires_info["label_type"] == "expression_list":
            requires_eval = adj(graph, val_id)
            for require in requires_eval:
                if graph.nodes[require].get("label_type") != "string":
                    continue
                dep_info = node_to_str(graph, require)
                yield new_conan_file_py_dep(
                    graph=graph,
                    node_id=val_id,
                    location=location,
                    is_dev=is_dev,
                    dep_info=dep_info,
                )
        elif requires_info["label_type"] == "list":
            arr_elements_ids = adj(graph, val_id)
            for elem_id in arr_elements_ids:
                elem_attrs = graph.nodes[elem_id]
                if elem_attrs["label_type"] not in {
                    "string",
                    "list",
                    "tuple",
                    "parenthesized_expression",
                }:
                    continue
                if elem_attrs["label_type"] in {"list", "tuple", "parenthesized_expression"}:
                    elem_id = adj(graph, elem_id)[1]  # noqa: PLW2901
                yield new_conan_file_py_dep(
                    graph=graph,
                    node_id=elem_id,
                    is_dev=is_dev,
                    location=location,
                )


def _resolve_deps(
    *,
    graph: Graph,
    conan_class_id: NId,
    location: Location,
    is_dev: bool,
    requires_attributes: dict[str, str],
) -> Iterator[Package]:
    class_block_id = graph.nodes[conan_class_id]["label_field_body"]

    for n_id in adj(graph, class_block_id):
        n_attrs = graph.nodes[n_id]
        if (
            n_attrs["label_type"] == "expression_statement"
            and (child := adj(graph, n_id)[0])
            and graph.nodes[child]["label_type"] == "assignment"
            and (left_id := graph.nodes[child].get("label_field_left"))
            and graph.nodes[left_id].get("label_text") == requires_attributes["requires"]
        ):
            yield from get_attr_requirements(
                graph=graph,
                attr_info=graph.nodes[child],
                location=location,
                is_dev=is_dev,
            )
        elif (
            n_attrs["label_type"] == "function_definition"
            and (name_id := graph.nodes[n_id].get("label_field_name"))
            and graph.nodes[name_id].get("label_text") == requires_attributes["requirements"]
        ):
            yield from get_method_requirements(
                graph=graph,
                method_id=n_id,
                location=location,
                is_dev=is_dev,
                requires_attributes=requires_attributes,
            )


def parse_conan_file_py(
    _: Resolver | None,
    __: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    packages: list[Package] = []
    content = reader.read_closer.read().encode("utf-8")
    if not content or not (graph := parse_ast_graph(content, "python")):
        return [], []
    conan_class_id = None
    for node in graph.nodes:
        if graph.nodes[node].get("label_type") != "class_definition":
            continue
        al_id = graph.nodes[node].get("label_field_superclasses")
        if al_id and any(
            graph.nodes[arg_id].get("label_text") == "ConanFile" for arg_id in adj(graph, al_id)
        ):
            conan_class_id = node
            break
    location = deepcopy(reader.location)
    if not conan_class_id:
        return [], []

    packages.extend(
        _resolve_deps(
            graph=graph,
            conan_class_id=conan_class_id,
            location=location,
            is_dev=False,
            requires_attributes={
                "requires": "requires",
                "requirements": "requirements",
                "method_requires": "self.requires",
            },
        ),
    )

    packages.extend(
        _resolve_deps(
            graph=graph,
            conan_class_id=conan_class_id,
            location=location,
            is_dev=True,
            requires_attributes={
                "requires": "tool_requires",
                "requirements": "build_requirements",
                "method_requires": "self.tool_requires",
            },
        ),
    )
    return packages, []



import logging
from copy import deepcopy
from typing import cast

from labels.model.file import Location, LocationReadCloser
from labels.model.indexables import IndexedDict, IndexedList, ParsedValue
from labels.model.package import Package
from labels.model.relationship import Relationship, RelationshipType
from labels.model.resolver import Resolver
from labels.parsers.cataloger.generic.parser import Environment
from labels.parsers.cataloger.php.model import PhpComposerInstalledEntry, PhpComposerLockEntry
from labels.parsers.cataloger.php.package import new_package_from_composer
from labels.parsers.collection.json import parse_json_with_tree_sitter

LOGGER = logging.getLogger(__name__)
EMPTY_LIST: IndexedList[ParsedValue] = IndexedList()


def _extract_packages(
    package_json: IndexedDict[str, ParsedValue],
    location: Location,
) -> list[Package]:
    packages = []
    if isinstance(package_json, IndexedDict) and "dev-package-names" in package_json:
        dev_packages = cast(IndexedList[ParsedValue], package_json["dev-package-names"])
    else:
        dev_packages = EMPTY_LIST

    packages_list = (
        package_json["packages"] if isinstance(package_json, IndexedDict) else package_json
    )

    for package in packages_list if isinstance(packages_list, IndexedList) else EMPTY_LIST:
        if not isinstance(package, IndexedDict):
            continue
        pkg_item = new_package_from_composer(package, deepcopy(location), is_dev=False)
        if pkg_item:
            pkg_item.is_dev = pkg_item.name in dev_packages
            packages.append(pkg_item)

    return packages


def _extract_relationships(packages: list[Package]) -> list[Relationship]:
    relationships = []
    for package in packages:
        if not isinstance(package.metadata, PhpComposerInstalledEntry | PhpComposerLockEntry):
            continue
        package_metadata = package.metadata.require
        dependencies = list(package_metadata.keys()) if package_metadata else []
        for dep_name in dependencies:
            package_dep = next((x for x in packages if x.name == dep_name), None)
            if package_dep:
                relationships.append(
                    Relationship(
                        from_=package,
                        to_=package_dep,
                        type=RelationshipType.DEPENDENCY_OF_RELATIONSHIP,
                    ),
                )

    return relationships


def parse_installed_json(
    _: Resolver | None,
    __: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    package_json = cast(
        IndexedDict[str, ParsedValue],
        parse_json_with_tree_sitter(reader.read_closer.read()),
    )

    packages = _extract_packages(package_json, reader.location)
    relationships = _extract_relationships(packages)

    return packages, relationships



import logging
import re
from copy import (
    deepcopy,
)

from labels.model.file import (
    DependencyType,
    Location,
    LocationReadCloser,
    Scope,
)
from labels.model.package import Language, Package, PackageType
from labels.model.relationship import (
    Relationship,
)
from labels.model.resolver import (
    Resolver,
)
from labels.parsers.cataloger.generic.parser import (
    Environment,
)
from labels.parsers.cataloger.ruby.package import (
    package_url,
)

LOGGER = logging.getLogger(__name__)

GEM_LOCK_DEP: re.Pattern[str] = re.compile(r"^\s{4}(?P<gem>[^\s]*)\s\([^\d]*(?P<version>.*)\)$")


def build_location(
    reader: LocationReadCloser,
    line_number: int,
) -> Location:
    location = deepcopy(reader.location)
    location.scope = Scope.PROD
    if location.coordinates:
        location.coordinates.line = line_number
        location.dependency_type = DependencyType.DIRECT
    return location


def parse_gemfile_lock(
    _: Resolver | None,
    __: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    packages: list[Package] = []
    relationships: list[Relationship] = []
    line_gem: bool = False

    for line_number, line in enumerate(
        reader.read_closer.read().splitlines(),
        1,
    ):
        if line.startswith("GEM"):
            line_gem = True
        elif line_gem:
            if matched := GEM_LOCK_DEP.match(line):
                pkg_name = matched.group("gem")
                pkg_version = matched.group("version")
                location = build_location(reader, line_number)

                packages.append(
                    Package(
                        name=pkg_name,
                        version=pkg_version,
                        type=PackageType.GemPkg,
                        locations=[location],
                        p_url=package_url(pkg_name, pkg_version),
                        metadata=None,
                        language=Language.RUBY,
                        licenses=[],
                        is_dev=False,
                    ),
                )
            elif not line:
                break

    return packages, relationships



import logging
from copy import (
    deepcopy,
)
from typing import (
    TypedDict,
    cast,
)

from pydantic import (
    BaseModel,
    ValidationError,
)

from labels.model.file import (
    Location,
    LocationReadCloser,
)
from labels.model.indexables import (
    IndexedDict,
    IndexedList,
    ParsedValue,
)
from labels.model.package import Language, Package, PackageType
from labels.model.relationship import (
    Relationship,
    RelationshipType,
)
from labels.model.resolver import (
    Resolver,
)
from labels.parsers.cataloger.generic.parser import (
    Environment,
)
from labels.parsers.cataloger.rust.utils import (
    package_url,
)
from labels.parsers.collection.toml import (
    parse_toml_with_tree_sitter,
)
from labels.utils.strings import format_exception

LOGGER = logging.getLogger(__name__)


class RustCargoLockEntry(BaseModel):
    name: str
    version: str
    source: str | None
    checksum: str | None
    dependencies: list[str]


class CargoLockEntry(TypedDict):
    name: str
    version: str
    source: str | None
    checksum: str | None
    dependencies: list[str]


class CargoLock(TypedDict):
    package: IndexedList[CargoLockEntry]


def _create_package(pkg: ParsedValue, location: Location) -> Package | None:
    if not isinstance(pkg, IndexedDict):
        return None
    name = str(pkg.get("name", "")) or None
    version = str(pkg.get("version", "")) or None

    if not name or not version:
        return None

    source = str(pkg.get("source", "")) or None
    dependencies: IndexedList[str] | None = cast(IndexedList[str] | None, pkg.get("dependencies"))
    checksum = str(pkg.get("checksum", "")) or None

    current_location = deepcopy(location)
    if current_location.coordinates:
        current_location.coordinates.line = pkg.get_key_position("version").start.line

    try:
        return Package(
            name=name,
            version=version,
            locations=[current_location],
            language=Language.RUST,
            licenses=[],
            p_url=package_url(name, version),
            type=PackageType.RustPkg,
            metadata=RustCargoLockEntry(
                name=name,
                version=version,
                source=source,
                dependencies=list(dependencies or []),
                checksum=checksum,
            ),
        )
    except ValidationError as ex:
        LOGGER.warning(
            "Malformed package. Required fields are missing or data types are incorrect.",
            extra={
                "extra": {
                    "exception": format_exception(str(ex)),
                    "location": current_location.path(),
                },
            },
        )
        return None


def _create_relationships(packages: list[Package]) -> list[Relationship]:
    relationships: list[Relationship] = []
    for pkg in packages:
        if isinstance(pkg.metadata, RustCargoLockEntry):
            relationships.extend(
                Relationship(
                    from_=pkg,
                    to_=dep,
                    type=RelationshipType.DEPENDENCY_OF_RELATIONSHIP,
                )
                for dep_name in pkg.metadata.dependencies
                if (dep := next((x for x in packages if x.name == dep_name), None))
            )
    return relationships


def parse_cargo_lock(
    _: Resolver | None,
    __: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    packages: list[Package] = []

    if not reader.location.coordinates:
        return packages, []

    _content = reader.read_closer.read()
    toml: IndexedDict[str, ParsedValue] = parse_toml_with_tree_sitter(
        _content,
    )

    toml_pkgs: ParsedValue = toml.get("package")
    if not isinstance(toml_pkgs, IndexedList):
        return [], []

    packages.extend(
        package
        for package in (_create_package(pkg, reader.location) for pkg in toml_pkgs)
        if package
    )

    relationships = _create_relationships(packages)

    return packages, relationships



import logging

from bs4 import BeautifulSoup, Tag

from labels.enrichers.debian.get import (
    DebianVersionInfo,
    get_deb_package_version_list,
    get_deb_snapshot,
)
from labels.enrichers.utils import infer_algorithm
from labels.model.package import Artifact, Digest, HealthMetadata, Package
from labels.parsers.cataloger.utils import extract_distro_info

LOGGER = logging.getLogger(__name__)


def is_stable_debian_version(version: str) -> bool:
    unstable_indicators = (
        "~rc",
        "~beta",
        "~alpha",
        "~dev",
        "~exp",  # Pre-release versions
        "+exp",  # Pre-release versions
        "+git",
        "~bpo",
        "~sid",  # Backports and sid/unstable
        "+nmu",  # Non-maintainer uploads
        "+next",
        "experimental",
        "testing",  # Explicit unstable/testing markers
    )

    return not any(indicator in version.lower() for indicator in unstable_indicators)


def get_tag_values(tag: Tag) -> tuple[str | list[str] | None, str | None]:
    tag_href: str | list[str] | None = None
    sha1_hash: str | None = None

    tag_href = tag.get("href")
    code_tag = tag.find_previous("code")
    if code_tag is not None:
        sha1_hash = code_tag.text
    return tag_href, sha1_hash


def _search_download_url(
    package: Package,
    arch: str | None = None,
) -> tuple[str, str | None] | None:
    html_download = get_deb_snapshot(package.name, package.version)
    if not html_download:
        return None
    parsed = BeautifulSoup(html_download, features="html.parser")
    tags: list[Tag] = parsed.find_all("a", href=True)
    tag_href: str | list[str] | None = None
    for tag in tags:
        if (
            tag.text.endswith(".deb")
            and package.name in tag.text
            and package.version in tag.text
            and (arch in tag.text if arch else True)
        ):
            tag_href, sha1_hash = get_tag_values(tag)
            break
    else:
        for tag in tags:
            if (
                tag.text.endswith(".deb")
                and package.name in tag.text
                and package.version in tag.text
            ):
                tag_href, sha1_hash = get_tag_values(tag)
                break
    if not tag_href:
        return None

    return f"https://snapshot.debian.org{tag_href}", sha1_hash


def _get_artifact(package: Package, arch: str | None) -> Artifact | None:
    download_url_item = _search_download_url(package, arch)
    if download_url_item:
        digest_value = download_url_item[1] or None
        return Artifact(
            url=download_url_item[0],
            integrity=Digest(
                algorithm=infer_algorithm(digest_value),
                value=digest_value,
            ),
        )
    return None


def _set_health_metadata(
    package: Package,
    versions_list: list[DebianVersionInfo] | None,
    arch: str | None,
) -> None:
    latest_version = None
    if versions_list:
        latest_version = next(
            (
                release["version"]
                for release in versions_list
                if is_stable_debian_version(release["version"])
            ),
            versions_list[0]["version"],
        )
    authors = package.metadata if hasattr(package.metadata, "maintainer") else None
    artifact = _get_artifact(package, arch)

    if not any([latest_version, authors, artifact]):
        return

    package.health_metadata = HealthMetadata(
        latest_version=latest_version,
        authors=authors.maintainer if authors else None,
        artifact=artifact,
    )


def complete_package(
    package: Package,
) -> Package:
    _, _, arch = extract_distro_info(package.p_url)
    versions_list = get_deb_package_version_list(package.name)
    _set_health_metadata(package, versions_list, arch)

    return package



from typing import TypedDict

import requests

from labels.config.cache import dual_cache

CRATES_ENDPOINT = "https://crates.io"


class Crate(TypedDict):
    max_stable_version: str
    updated_at: str


class Publisher(TypedDict):
    name: str


class Version(TypedDict):
    checksum: str
    dl_path: str
    license: str
    num: str
    published_by: Publisher


class CargoPackage(TypedDict):
    crate: Crate
    versions: list[Version]


@dual_cache
def get_cargo_package(package_name: str) -> CargoPackage | None:
    url = f"{CRATES_ENDPOINT}/api/v1/crates/{package_name}"
    response = requests.get(url, timeout=30)
    if response.status_code == 200:
        result: CargoPackage = response.json()
        return result

    return None



import logging
from collections.abc import Callable
from typing import cast

from pydantic import ValidationError

import labels.enrichers.alpine.complete as enrich_alpine
import labels.enrichers.dart.complete as enrich_dart
import labels.enrichers.debian.complete as enrich_debian
import labels.enrichers.dotnet.complete as enrich_dotnet
import labels.enrichers.elixir.complete as enrich_elixir
import labels.enrichers.golang.complete as enrich_go
import labels.enrichers.java.complete as enrich_java
import labels.enrichers.javascript.complete as enrich_js
import labels.enrichers.php.complete as enrich_php
import labels.enrichers.python.complete as enrich_python
import labels.enrichers.ruby.complete as enrich_ruby
import labels.enrichers.rust.complete as enrich_rust
from labels.advisories import images as images_advisories
from labels.advisories import roots as roots_advisories
from labels.model.advisories import Advisory
from labels.model.file import Location
from labels.model.package import HealthMetadata, Language, Package, PackageType
from labels.parsers.cataloger.utils import extract_distro_info
from labels.utils.strings import format_exception

LOGGER = logging.getLogger(__name__)


ROOT_TYPES: set[PackageType] = {
    PackageType.NpmPkg,
    PackageType.DartPubPkg,
    PackageType.DotnetPkg,
    PackageType.JavaPkg,
    PackageType.PhpComposerPkg,
    PackageType.PythonPkg,
    PackageType.GemPkg,
    PackageType.RustPkg,
    PackageType.GoModulePkg,
    PackageType.HexPkg,
    PackageType.ConanPkg,
    PackageType.GithubActionPkg,
    PackageType.SwiftPkg,
}

IMAGES_TYPES: set[PackageType] = {
    PackageType.DebPkg,
    PackageType.ApkPkg,
}

COMPLETION_MAP: dict[PackageType, Callable[[Package], Package]] = {
    PackageType.NpmPkg: enrich_js.complete_package,
    PackageType.DartPubPkg: enrich_dart.complete_package,
    PackageType.DotnetPkg: enrich_dotnet.complete_package,
    PackageType.JavaPkg: enrich_java.complete_package,
    PackageType.PhpComposerPkg: enrich_php.complete_package,
    PackageType.PythonPkg: enrich_python.complete_package,
    PackageType.GemPkg: enrich_ruby.complete_package,
    PackageType.RustPkg: enrich_rust.complete_package,
    PackageType.GoModulePkg: enrich_go.complete_package,
    PackageType.HexPkg: enrich_elixir.complete_package,
    PackageType.DebPkg: enrich_debian.complete_package,
    PackageType.ApkPkg: enrich_alpine.complete_package,
}

ALLOWED_TYPE = dict[
    str,
    str
    | Language
    | list[str]
    | list[Location]
    | PackageType
    | list[Advisory]
    | list[Package]
    | HealthMetadata
    | bool
    | object
    | None,
]


def update_root_advisories(package: Package) -> list[Advisory]:
    if pkg_platform := package.type.get_platform_value():
        return roots_advisories.get_vulnerabilities(
            pkg_platform.lower(),
            package.name,
            package.version,
        )
    return []


def update_image_advisories(package: Package) -> list[Advisory]:
    distro_id, distro_version, _ = extract_distro_info(package.p_url)
    distro_version = (
        "v" + ".".join(str(distro_version).split(".")[0:2])
        if package.type == PackageType.ApkPkg
        else str(distro_version)
    )

    return images_advisories.get_vulnerabilities(
        str(distro_id),
        package.name,
        package.version,
        distro_version,
    )


def add_package_advisories(package: Package) -> list[Advisory] | None:
    try:
        pkg_advisories = []
        if package.type in ROOT_TYPES:
            pkg_advisories = update_root_advisories(package)
        if package.type in IMAGES_TYPES:
            pkg_advisories = update_image_advisories(package)
    except ValidationError as ex:
        LOGGER.exception(
            "Unable to complete package advisories",
            extra={
                "extra": {
                    "exception": format_exception(str(ex)),
                    "location": package.locations,
                    "package_type": package.type,
                },
            },
        )
        return None
    return pkg_advisories


def complete_package_advisories_only(package: Package) -> Package:
    if pkg_advisories := add_package_advisories(package):
        package.advisories = pkg_advisories
    return package


def complete_package(package: Package) -> Package:
    try:
        if pkg_advisories := add_package_advisories(package):
            package.advisories = pkg_advisories

        completed_package = None
        if package.type in COMPLETION_MAP:
            try:
                completed_package = COMPLETION_MAP[package.type](package)

            except Exception as ex:
                LOGGER.exception(
                    "Unable to complete package",
                    extra={
                        "extra": {
                            "exception": format_exception(str(ex)),
                            "location": package.locations,
                            "package_type": package.type,
                        },
                    },
                )
                return package
        if completed_package:
            completed_package.model_validate(
                cast("ALLOWED_TYPE", package.__dict__),
            )
    except ValidationError:
        LOGGER.warning(
            "Malformed package metadata completion. "
            "Required fields are missing or data types are incorrect.",
        )
        return package

    return completed_package if completed_package is not None else package



import logging
from copy import deepcopy

from bs4 import BeautifulSoup
from packageurl import PackageURL
from pydantic import ValidationError

from labels.model.file import DependencyType, LocationReadCloser
from labels.model.package import Language, Package, PackageType
from labels.model.relationship import Relationship
from labels.model.resolver import Resolver
from labels.parsers.cataloger.generic.parser import Environment
from labels.utils.strings import format_exception

LOGGER = logging.getLogger(__name__)


def parse_dotnet_pkgs_config(
    _resolver: Resolver | None,
    _env: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    root = BeautifulSoup(reader.read_closer.read(), features="html.parser")
    packages = []

    for pkg in root.find_all("package", recursive=True):
        name: str | None = pkg.get("id")
        version: str | None = pkg.get("version")

        if not name or not version:
            continue

        line = pkg.sourceline
        location = deepcopy(reader.location)
        if location.coordinates:
            location.coordinates.line = line
            location.dependency_type = DependencyType.DIRECT

        try:
            packages.append(
                Package(
                    name=name,
                    version=version,
                    locations=[location],
                    language=Language.DOTNET,
                    licenses=[],
                    type=PackageType.DotnetPkg,
                    metadata=None,
                    p_url=PackageURL(
                        type="nuget",
                        namespace="",
                        name=name,
                        version=version,
                        qualifiers={},
                        subpath="",
                    ).to_string(),
                ),
            )
        except ValidationError as ex:
            LOGGER.warning(
                "Malformed package. Required fields are missing or data types are incorrect.",
                extra={
                    "extra": {
                        "exception": format_exception(str(ex)),
                        "location": location.path(),
                    },
                },
            )
            continue

    return packages, []



from typing import NotRequired, TypedDict

from labels.enrichers.api_interface import make_get


class PubSpec(TypedDict):
    author: NotRequired[str]


class PubPackageVersion(TypedDict):
    archive_sha256: str
    archive_url: str
    published: str
    pubspec: PubSpec
    version: str


class PubPackage(TypedDict):
    latest: PubPackageVersion
    name: str
    versions: list[PubPackageVersion]


def get_pub_package(package_name: str) -> PubPackage | None:
    url = f"https://pub.dev/api/packages/{package_name}"
    return make_get(url, timeout=30, headers={"Accept": "gzip"})



from typing import NotRequired, TypedDict

import requests

from labels.config.cache import dual_cache


class Author(TypedDict):
    name: str
    homepage: str
    email: NotRequired[str]


class Source(TypedDict):
    url: str
    type: str
    reference: str


class Dist(TypedDict):
    url: str
    type: str
    shasum: str
    reference: str


class PackagistPackageInfo(TypedDict, total=False):
    name: str
    description: str
    keywords: list[str]
    homepage: str
    version: str
    version_normalized: str
    license: list[str]
    authors: list[Author]
    source: Source
    dist: Dist
    type: str
    support: dict[str, str]
    funding: list[str]
    time: str
    autoload: dict[str, list[str]]
    extra: dict[str, dict[str, str]]
    require: dict[str, str]
    require_dev: dict[str, str]
    suggest: dict[str, str]
    conflict: dict[str, str]


class PackagistResponse(TypedDict, total=False):
    minified: str
    packages: dict[str, list[PackagistPackageInfo]]


def get_package_data(
    package_name: str,
    base_url: str = "https://repo.packagist.org",
) -> PackagistResponse | None:
    url = f"{base_url}/p2/{package_name}.json"
    response = requests.get(url, timeout=30)

    if response.status_code == 200:
        return response.json()
    return None


def find_version(
    package_versions: list[PackagistPackageInfo],
    version: str | None,
) -> PackagistPackageInfo | None:
    if not version:
        return package_versions[0] if package_versions else None

    return next(
        (
            version_data
            for version_data in package_versions
            if version_data.get("version") == version
        ),
        None,
    )


@dual_cache
def get_composer_package(
    package_name: str,
    version: str | None = None,
) -> PackagistPackageInfo | None:
    response_data = get_package_data(package_name)
    if not response_data:
        return None

    package_versions = response_data.get("packages", {}).get(package_name, [])
    return find_version(package_versions, version)



from typing import Any

from retry_requests import retry

from labels.config.cache import dual_cache


@dual_cache
def make_get(url: str, *, content: bool = False, **kwargs: Any) -> Any | None:
    response = retry().get(url, timeout=kwargs.pop("timeout", 30), **kwargs)
    if response.status_code != 200:
        return None
    if content:
        return response.content.decode("utf-8")

    return response.json()



import struct

from labels.config.logger import LOGGER


def htonl(val: int) -> int:
    try:
        # Convert from little-endian (host byte order)
        # to big-endian (network byte order)
        return struct.unpack(">i", struct.pack("<i", val))[0]
    except struct.error as exc:
        LOGGER.error("Failed to convert integer: %s", exc)
        return 0


def htonlu(val: int) -> int:
    try:
        # Convert from little-endian (host byte order)
        # to big-endian (network byte order)
        return struct.unpack(">I", struct.pack("<I", val))[0]
    except struct.error as exc:
        LOGGER.error("Failed to convert unsigned integer: %s", exc)
        return 0



from labels.enrichers.python.get import PyPIResponse, get_pypi_package
from labels.enrichers.utils import infer_algorithm
from labels.model.package import Artifact, Digest, HealthMetadata, Package
from labels.utils.licenses.validation import validate_licenses


def _get_artifact(package: Package, current_package: PyPIResponse) -> Artifact | None:
    url = next(
        (x for x in current_package["urls"] if x["url"].endswith(".tar.gz")),
        None,
    )

    digest_value: str | None = url.get("digests", {}).get("sha256") or None if url else None

    return Artifact(
        url=url["url"] if url else f"https://pypi.org/pypi/{package.name}",
        integrity=Digest(
            algorithm=infer_algorithm(digest_value),
            value=digest_value,
        ),
    )


def _get_authors(pypi_package: PyPIResponse) -> str | None:
    package_info = pypi_package["info"]
    author: str | None = None
    package_author = package_info["author"]
    author_email = package_info.get("author_email")
    if package_author:
        author = package_author
    if not author and author_email:
        author = author_email
    if author and author_email and author_email not in author:
        author = f"{author} <{author_email}>"
    return author


def _set_health_metadata(
    package: Package,
    pypi_package: PyPIResponse,
    current_package: PyPIResponse | None,
) -> None:
    info = pypi_package.get("info")
    if not isinstance(info, dict):
        return
    pypi_package_version = info.get("version")
    releases = pypi_package.get("releases")
    if not isinstance(releases, dict):
        releases = {}

    upload_time = releases.get(pypi_package_version) if pypi_package_version else []
    if not isinstance(upload_time, list):
        upload_time = []

    latest_version_created_at: str | None = None
    if upload_time:
        first_release = upload_time[0]
        time_value = first_release.get("upload_time_iso_8601")
        if isinstance(time_value, str):
            latest_version_created_at = time_value

    package.health_metadata = HealthMetadata(
        latest_version=pypi_package_version,
        latest_version_created_at=latest_version_created_at,
        authors=_get_authors(pypi_package),
        artifact=_get_artifact(package, current_package) if current_package else None,
    )


def _update_licenses(pypi_package: PyPIResponse, package: Package) -> None:
    info = pypi_package.get("info")
    if not isinstance(info, dict):
        return
    licenses = info.get("license")
    if licenses:
        package.licenses = validate_licenses([licenses])


def complete_package(package: Package) -> Package:
    pypi_package = get_pypi_package(package.name)
    if not pypi_package:
        return package

    current_package = get_pypi_package(package.name, package.version)

    _set_health_metadata(package, pypi_package, current_package)

    _update_licenses(pypi_package, package)

    return package



import logging
import re

from labels.parsers.cataloger.java.model import (
    JavaPomProperties,
)

LOGGER = logging.getLogger(__name__)


def parse_pom_properties(path: str, file_content: str) -> JavaPomProperties | None:
    prop_map = {}

    for raw_line in file_content.splitlines():
        line = raw_line.strip()
        # Skip empty lines and comments
        if line == "" or line.lstrip().startswith("#"):
            continue

        # Find the first occurrence of ':' or '='
        idx = next((i for i in range(len(line)) if line[i] in ":="), -1)
        if idx == -1:
            LOGGER.error("Unable to split pom.properties line into key-value pairs: %s", line)
            continue

        key = line[:idx].strip()
        value = line[idx + 1 :].strip()
        prop_map[key] = value

    # Convert the dictionary to a JavaPomProperties object
    props = JavaPomProperties(path=path)
    for raw_key, value in prop_map.items():
        key = re.sub(r"(?<!^)(?=[A-Z])", "_", raw_key).lower()
        if hasattr(props, key):
            setattr(props, key, value)
        else:
            props.extra[key] = value

    return props



import logging
import re
from copy import deepcopy

from bs4 import BeautifulSoup
from packageurl import PackageURL
from pydantic import ValidationError

from labels.model.file import DependencyType, LocationReadCloser
from labels.model.package import Language, Package, PackageType
from labels.model.relationship import Relationship
from labels.model.resolver import Resolver
from labels.parsers.cataloger.generic.parser import Environment
from labels.utils.strings import format_exception

LOGGER = logging.getLogger(__name__)
PACKAGE = re.compile(r".+\\packages\\(?P<package_info>[^\s\\]*)\\.+")
DEP_INFO = re.compile(r"(?P<package_name>.*?)\.(?P<version>\d+[^\s]*)$")


def _get_version(include_info: list[str]) -> str | None:
    return next(
        (
            pkg_info.lstrip("Version=")
            for pkg_info in include_info
            if pkg_info.startswith("Version=")
        ),
        None,
    )


def format_package(
    name: str,
    version: str,
    line: int,
    reader: LocationReadCloser,
) -> Package | None:
    location = deepcopy(reader.location)
    if location.coordinates:
        location.coordinates.line = line
        location.dependency_type = DependencyType.DIRECT

    try:
        return Package(
            name=name,
            version=version,
            locations=[location],
            language=Language.DOTNET,
            licenses=[],
            type=PackageType.DotnetPkg,
            metadata=None,
            p_url=PackageURL(
                type="nuget",
                namespace="",
                name=name,
                version=version,
                qualifiers={},
                subpath="",
            ).to_string(),
        )
    except ValidationError as ex:
        LOGGER.warning(
            "Malformed package. Required fields are missing or data types are incorrect.",
            extra={
                "extra": {
                    "exception": format_exception(str(ex)),
                    "location": location.path(),
                },
            },
        )
    return None


def _format_csproj_reference_deps(root: BeautifulSoup, reader: LocationReadCloser) -> list[Package]:
    packages = []
    for pkg in root.find_all("reference", recursive=True):
        if dll_path := pkg.find("hintpath"):
            package = PACKAGE.match(dll_path.text)
            if (
                package
                and (pkg_info := DEP_INFO.match(package.group("package_info")))
                and (
                    pkg := format_package(
                        str(pkg_info.group("package_name").lower()),
                        str(pkg_info.group("version")),
                        dll_path.sourceline,
                        reader,
                    )
                )
            ):
                packages.append(pkg)
        elif (include := pkg.get("include")) and (
            include_info := include.replace(" ", "").split(",")
        ):
            version = _get_version(include_info)
            if version and (
                pkg := format_package(
                    str(include_info[0]).strip(),
                    str(version),
                    pkg.sourceline,
                    reader,
                )
            ):
                packages.append(pkg)

    return packages


def parse_csproj(
    _resolver: Resolver | None,
    _env: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    packages: list[Package] = []
    root = BeautifulSoup(reader.read_closer.read(), features="html.parser")
    packages = [
        package
        for pkg in root.find_all("packagereference", recursive=True)
        if (
            (pkg_name := pkg.get("include"))
            and (version := pkg.get("version"))
            and (package := format_package(pkg_name, version, pkg.sourceline, reader))
        )
    ] + _format_csproj_reference_deps(root, reader)

    return packages, []



import logging
import re
from contextlib import suppress
from copy import deepcopy
from pathlib import Path
from typing import TextIO

from bs4 import BeautifulSoup, NavigableString, Tag
from pydantic import BaseModel, ValidationError

from labels.model.file import DependencyType, Location, LocationReadCloser
from labels.model.package import Language, Package, PackageType
from labels.model.relationship import Relationship
from labels.model.resolver import Resolver
from labels.parsers.cataloger.generic.parser import Environment
from labels.parsers.cataloger.java.maven_repo_utils import recursively_find_versions_from_parent_pom
from labels.parsers.cataloger.java.model import (
    JavaArchive,
    JavaPomParent,
    JavaPomProject,
    JavaPomProperties,
    PomContext,
)
from labels.parsers.cataloger.java.package import package_url
from labels.utils.strings import format_exception

LOGGER = logging.getLogger(__name__)


def get_pom_project(content: TextIO) -> Tag | None:
    if (
        (root := BeautifulSoup(content, features="html.parser"))
        and (pom_project := root.project)
        and (xmlns := root.project.get("xmlns"))
        and str(xmlns) == "http://maven.apache.org/POM/4.0.0"
    ):
        return pom_project
    return None


class ParsedPomProject(BaseModel):
    java_pom_project: JavaPomProject
    licenses: list[str]


def extract_bracketed_text(item: str) -> str:
    match = re.search(r"\$\{([^}]+)\}", item)
    if match:
        return match.group(1)
    return ""


def _get_text(parent: Tag | NavigableString, name: str) -> str | None:
    element = parent.find_next(name)
    if element:
        return element.get_text()
    return None


def _get_parent_info(project: Tag) -> dict[str, str] | None:
    if (
        (parent := project.find_next("parent"))
        and (parent_groupid_node := parent.find_next("groupid"))
        and (parent_artifactid_node := parent.find_next("artifactid"))
        and (parent_version_node := parent.find_next("version"))
    ):
        return {
            "group": parent_groupid_node.get_text(),
            "artifact": parent_artifactid_node.get_text(),
            "version": parent_version_node.get_text(),
        }
    return None


def _is_parent_pom(root_pom_project: Tag, parent_info: dict[str, str]) -> bool:
    group = root_pom_project.find("groupid", recursive=False)
    artifact = root_pom_project.find("artifactid", recursive=False)
    version = root_pom_project.find("version", recursive=False)
    if not (group and artifact and version):
        return False
    return (
        group.get_text() == parent_info["group"]
        and artifact.get_text() == parent_info["artifact"]
        and version.get_text() == parent_info["version"]
    )


def _is_module_parent(parent_pom: Tag, pom_module: str) -> bool:
    base_module = Path(pom_module).parent.name
    for modules in parent_pom.find_all("modules"):
        for module in modules.find_all("module"):
            mod_name = module.get_text()
            if mod_name == base_module:
                return True
    return False


def _get_properties(project_pom: Tag) -> dict[str, str]:
    return {
        _property.name.lower(): _property.get_text()
        for properties in project_pom.find_all("properties", limit=2)
        for _property in properties.children
        if isinstance(_property, Tag)
    }


def _add_properties_vars(properties_vars: dict[str, str], root_pom: Tag) -> None:
    properties = _get_properties(root_pom)
    properties_vars.update(properties)


def _get_deps_management(pom_tree: Tag) -> dict[str, str]:
    deps_info: dict[str, str] = {}
    for manage in pom_tree.find_all("dependencymanagement"):
        for dependency in manage.find_all("dependency", recursive=True):
            if not (dependency.groupid and dependency.artifactid and dependency.version):
                continue

            group = dependency.groupid.get_text()
            artifact = dependency.artifactid.get_text()
            version = dependency.version.get_text()
            deps_info[f"{group}:{artifact}"] = version
    return deps_info


def _evaluate_pom_files_in_project(
    resolver: Resolver,
    parent_info: dict[str, str],
    current_pom_file_path: str,
    current_pom_project: Tag,
) -> tuple[dict[str, str], dict[str, str]]:
    properties_vars: dict[str, str] = {}
    manage_deps: dict[str, str] = {}
    for pom_file_location in resolver.files_by_glob("**/pom.xml", "pom.xml"):
        content = resolver.file_contents_by_location(pom_file_location)
        if (
            content
            and (pom_project := get_pom_project(content))
            and _is_parent_pom(pom_project, parent_info)
            and _is_module_parent(pom_project, current_pom_file_path)
        ):
            _add_properties_vars(properties_vars, pom_project)
            manage_deps.update(_get_deps_management(pom_project))
            break

    manage_deps.update(_get_deps_management(current_pom_project))
    return properties_vars, manage_deps


def is_version_correct(version: str | None) -> bool:
    return version is not None and not version.startswith("${")


def update_location_with_dependency_info(location: Location, dependency: Tag) -> None:
    if location.coordinates:
        location.coordinates.line = (
            dependency.version.sourceline if dependency.version else dependency.sourceline
        )
        location.dependency_type = DependencyType.DIRECT


def new_package_from_pom_xml(
    project: Tag,
    dependency: Tag,
    location: Location,
    context: PomContext,
) -> Package | None:
    name = _get_text(dependency, "artifactid")
    if not name:
        return None

    group_id = _get_text(dependency, "groupid")
    full_name = f"{group_id}:{name}" if group_id else name

    java_archive = JavaArchive(
        pom_properties=JavaPomProperties(
            group_id=group_id,
            artifact_id=name,
            version=_get_text(dependency, "version"),
        ),
    )

    version = resolve_version(
        dependency,
        project,
        context,
        java_archive,
        full_name,
    )
    if not is_version_correct(version):
        return None

    update_location_with_dependency_info(location, dependency)

    try:
        return Package(
            name=full_name,
            version=str(version),
            licenses=[],
            locations=[location],
            language=Language.JAVA,
            type=PackageType.JavaPkg,
            metadata=java_archive,
            p_url=package_url(name, str(version), java_archive),
        )
    except ValidationError as ex:
        log_validation_error(ex, location)
        return None


def resolve_version(
    dependency: Tag,
    project: Tag,
    context: PomContext,
    java_archive: JavaArchive,
    full_name: str,
) -> str | None:
    version = dependency.version.get_text() if dependency.version else None

    if version and version.startswith("${"):
        version = resolve_property_version(version, project, context.parent_version_properties)

    if not version and context.parent_info:
        version = resolve_parent_version(java_archive, context.parent_info)

    if not version and context.manage_deps and context.parent_version_properties:
        version = resolve_managed_version(
            context.manage_deps,
            context.parent_version_properties,
            full_name,
        )

    return version


def resolve_property_version(
    version: str,
    project: Tag,
    parent_version_properties: dict[str, str] | None,
) -> str | None:
    property_name = extract_bracketed_text(version)
    property_node = project.find_next(property_name)
    if property_node:
        version_text = property_node.get_text()
        if version_text and not version_text.startswith("${"):
            return version_text

    if parent_version_properties:
        return parent_version_properties.get(property_name)
    return None


def resolve_parent_version(java_archive: JavaArchive, parent_info: dict[str, str]) -> str | None:
    if (
        java_archive.pom_properties
        and java_archive.pom_properties.group_id
        and java_archive.pom_properties.artifact_id
    ):
        return recursively_find_versions_from_parent_pom(
            group_id=java_archive.pom_properties.group_id,
            artifact_id=java_archive.pom_properties.artifact_id,
            parent_group_id=parent_info["group"],
            parent_artifact_id=parent_info["artifact"],
            parent_version=parent_info["version"],
        )
    return None


def resolve_managed_version(
    manage_deps: dict[str, str],
    parent_version_properties: dict[str, str],
    full_name: str,
) -> str | None:
    managed_version = manage_deps.get(full_name)
    if managed_version:
        if not managed_version.startswith("${"):
            return managed_version
        return parent_version_properties.get(extract_bracketed_text(managed_version))
    return None


def log_validation_error(ex: ValidationError, location: Location) -> None:
    LOGGER.warning(
        "Malformed package. Required fields are missing or data types are incorrect.",
        extra={
            "extra": {
                "exception": format_exception(str(ex)),
                "location": location.path(),
            },
        },
    )


def parse_pom_xml(
    resolver: Resolver | None,
    _env: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    root = None
    with suppress(UnicodeError):
        try:
            root = BeautifulSoup(reader.read_closer, features="html.parser")
        except AssertionError:
            return [], []

    if not root:
        return [], []

    pkgs = []
    if (
        (project := root.project)
        and str(project.get("xmlns")) == "http://maven.apache.org/POM/4.0.0"
        and (dependencies := project.find("dependencies", recursive=False))
        and isinstance(dependencies, Tag)
    ):
        parent_info = _get_parent_info(project)
        parent_version_properties = None
        manage_deps = None
        if resolver and parent_info:
            parent_version_properties, manage_deps = _evaluate_pom_files_in_project(
                resolver,
                parent_info,
                str(reader.location.access_path),
                project,
            )
        else:
            parent_version_properties = _get_properties(root)
            manage_deps = _get_deps_management(root)

        context = PomContext(
            parent_info=parent_info,
            parent_version_properties=parent_version_properties,
            manage_deps=manage_deps,
        )
        for dependency in dependencies.find_all("dependency"):
            pkg = new_package_from_pom_xml(
                project,
                dependency,
                deepcopy(reader.location),
                context,
            )
            if pkg:
                pkgs.append(pkg)
    return pkgs, []


def decode_pom_xml(content: str) -> Tag:
    return BeautifulSoup(content, features="html.parser")


def pom_parent(parent: Tag | None) -> JavaPomParent | None:
    if not parent:
        return None

    group_id = _get_text(parent, "groupId")
    artifact_id = _get_text(parent, "artifactId")
    version = _get_text(parent, "version")

    if (
        not group_id
        or not artifact_id
        or not version
        or not group_id.strip()
        or not artifact_id.strip()
        or not version.strip()
    ):
        return None

    return JavaPomParent(
        group_id=group_id,
        artifact_id=artifact_id,
        version=version,
    )


def parse_pom_xml_project(
    path: str,
    reader: str,
    _location: Location,
) -> ParsedPomProject | None:
    project = BeautifulSoup(reader, features="xml").project
    if not project:
        return None
    return new_pom_project(path, project, _location)


def _find_direct_child(parent: Tag, tag: str) -> Tag | None:
    return next(
        (child for child in parent.find_all(tag, recursive=False) if child.parent == parent),
        None,
    )


def new_pom_project(
    path: str,
    project: Tag,
    _location: Location,
) -> ParsedPomProject:
    artifact_id = _safe_string(_find_direct_child(project, "artifactId"))
    name = _safe_string(_find_direct_child(project, "name"))
    project_url = _safe_string(_find_direct_child(project, "url"))

    licenses: list[str] = []
    if project.licenses:
        for license_ in project.licenses.find_all("license"):
            license_name: str | None = None
            license_url: str | None = None
            if name_node := license_.find_next("name"):
                license_name = name_node.get_text()
            elif url_node := license_.find_next("url"):
                license_url = url_node.get_text()

            if not license_name and not license_url:
                continue

            license_value = license_name or license_url

            if license_value is not None:
                licenses.append(license_value)

    return ParsedPomProject(
        java_pom_project=JavaPomProject(
            path=path,
            parent=pom_parent(_find_direct_child(project, "parent")),
            group_id=_safe_string(_find_direct_child(project, "groupId")),
            artifact_id=artifact_id,
            version=_safe_string(_find_direct_child(project, "version")),
            name=name,
            description=_safe_string(
                _find_direct_child(project, "description"),
            ),
            url=project_url,
        ),
        licenses=licenses,
    )


def _safe_string(value: Tag | None) -> str:
    if not value:
        return ""
    return value.get_text()



from labels.enrichers.api_interface import make_get


def normalize_architecture(arch: str | None) -> str | None:
    if not arch:
        return None

    arch_mapping = {
        "x86": "x86",
        "x86_64": "x86_64",
        "amd64": "x86_64",
        "i386": "x86",
        "i686": "x86",
        "arm": "arm",
        "armhf": "arm",
        "armv7": "arm",
        "arm64": "aarch64",
        "aarch64": "aarch64",
        "ppc64le": "ppc64le",
        "s390x": "s390x",
    }

    return arch_mapping.get(arch.lower(), arch)


def format_distro_version(version: str | None) -> str:
    if not version:
        return "edge"

    parts = version.split(".")
    if len(parts) == 3:
        return f"v{parts[0]}.{parts[1]}"

    return version


def get_package_versions_html(
    name: str,
    distro_version: str | None = None,
    arch: str | None = None,
) -> str | None:
    normalized_arch = normalize_architecture(arch)
    if normalized_arch is None:
        return None

    branch = format_distro_version(distro_version)

    return make_get(
        "https://pkgs.alpinelinux.org/packages",
        params={
            "name": name,
            "branch": branch,
            "repo": "",
            "arch": normalized_arch,
            "maintainer": "",
        },
        content=True,
    )



import functools
import hashlib
import os
import pickle  # nosec
import socket
from collections.abc import Callable
from contextlib import suppress
from typing import Any, TypeVar, cast

import diskcache
import diskcache.core
import redis
from diskcache import Lock
from platformdirs import user_cache_dir

REDIS_CACHE_ENDPOINT = os.environ.get("CACHE_URL")
REDIS_CLIENT = None

TVar = TypeVar("TVar")
TFun = TypeVar("TFun", bound=Callable[..., Any])

if REDIS_CACHE_ENDPOINT:
    try:
        host, _port = REDIS_CACHE_ENDPOINT.split(":", maxsplit=1)
        port = int(_port)

        # Try to connect to redis endpoint
        socket.create_connection((host, port), timeout=2)

        REDIS_CLIENT = redis.StrictRedis(
            host=host,
            port=port,
            ssl=True,
            username="readwrite",
            password=os.environ.get("CACHE_USER_WRITE_PASSWORD"),
        )
    except (OSError, ValueError):
        REDIS_CLIENT = None

DISK_CACHE = diskcache.Cache(user_cache_dir("fluid-labels", "fluidattacks"))


# 4 weeks
TTL = 604800 * 4


def make_hashable(item: Any) -> str:
    serialized_object = pickle.dumps(item)

    return hashlib.sha256(serialized_object).hexdigest()


def generate_cache_key(func: Any, args: Any, kwargs: dict[Any, Any]) -> str:
    key = f"{func.__module__}.{func.__name__}-"
    key += str(make_hashable((args, kwargs)))
    return hashlib.sha256(key.encode()).hexdigest()


def dual_cache(func: TVar) -> TVar:
    _func = cast(Callable[..., Any], func)

    @functools.wraps(_func)
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        if os.environ.get("RUNNING_TESTS", "true").lower() == "true":
            return _func(*args, **kwargs)

        cache_key = generate_cache_key(_func, args, kwargs)

        # Try to recover from disk cache
        if value := DISK_CACHE.get(cache_key):
            return value
        # Try to recover from Redis cache (ElastiCache) if available
        if REDIS_CLIENT:
            cached_result = REDIS_CLIENT.get(cache_key)

            if cached_result:
                result = pickle.loads(  # noqa: S301
                    cached_result,
                )
                with Lock(DISK_CACHE, "dual_cache"):
                    DISK_CACHE.set(
                        cache_key,
                        result,
                        expire=TTL,
                        retry=True,
                    )  # Save to disk with TTL

                return result

        # Run the function and store the result in both caches
        result = _func(*args, **kwargs)
        with suppress(diskcache.core.Timeout), Lock(DISK_CACHE, "dual_cache"):
            DISK_CACHE.set(cache_key, result, expire=TTL, retry=True)  # Save to disk with TTL
        if REDIS_CLIENT:
            REDIS_CLIENT.setex(cache_key, TTL, pickle.dumps(result))  # Save to Redis with TTL
        return result

    return cast(TVar, wrapper)



from packageurl import PackageURL


def package_url(name: str, version: str) -> str:
    return PackageURL(  # type: ignore
        type="hex",
        namespace="",
        name=name,
        version=version,
        qualifiers=None,
        subpath="",
    ).to_string()



import logging

from pydantic import ValidationError

from labels.model.file import Location, LocationReadCloser
from labels.model.indexables import IndexedDict, IndexedList, ParsedValue
from labels.model.package import Language, Package, PackageType
from labels.model.relationship import Relationship, RelationshipType
from labels.model.resolver import Resolver
from labels.parsers.cataloger.generic.parser import Environment
from labels.parsers.cataloger.python.model import PythonRequirementsEntry
from labels.parsers.cataloger.python.package import package_url
from labels.parsers.collection import toml
from labels.utils.strings import format_exception

LOGGER = logging.getLogger(__name__)


def _get_location(location: Location, sourceline: int) -> Location:
    if location.coordinates:
        c_upd = {"line": sourceline}
        l_upd = {"coordinates": location.coordinates.model_copy(update=c_upd)}
        return location.model_copy(update=l_upd)
    return location


def parse_poetry_lock(
    _resolver: Resolver | None,
    _env: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    _content = reader.read_closer.read()

    toml_content: IndexedDict[str, ParsedValue] = toml.parse_toml_with_tree_sitter(_content)

    packages = _parse_packages(toml_content, reader)
    relationships = _parse_relationships(toml_content, packages)

    return packages, relationships


def _parse_packages(
    toml_content: IndexedDict[str, ParsedValue],
    reader: LocationReadCloser,
) -> list[Package]:
    packages = []
    toml_pkgs = toml_content.get("package")
    if not isinstance(toml_pkgs, IndexedList):
        return []
    for package in toml_pkgs:
        if not isinstance(package, IndexedDict):
            continue
        name: str = str(package.get("name", ""))
        version: str = str(package.get("version", ""))

        if not name or not version:
            continue
        p_url = package_url(name, version, package)  # type: ignore

        location = (
            _get_location(reader.location, package.get_key_position("version").start.line)
            if isinstance(package, IndexedDict)
            else reader.location
        )

        try:
            packages.append(
                Package(
                    name=name,
                    version=version,
                    found_by=None,
                    locations=[location],
                    language=Language.PYTHON,
                    p_url=p_url,
                    metadata=PythonRequirementsEntry(
                        name=name,
                        extras=[],
                        markers=p_url,
                    ),
                    licenses=[],
                    type=PackageType.PythonPkg,
                ),
            )
        except ValidationError as ex:
            LOGGER.warning(
                "Malformed package. Required fields are missing or data types are incorrect.",
                extra={
                    "extra": {
                        "exception": format_exception(str(ex)),
                        "location": location.path(),
                    },
                },
            )
            continue

    return packages


def _get_dependencies(
    package: ParsedValue,
    packages: list[Package],
) -> tuple[Package | None, IndexedDict[str, ParsedValue]] | None:
    if not isinstance(package, IndexedDict):
        return None
    _pkg = next((pkg for pkg in packages if pkg.name == package["name"]), None)
    deps = package.get("dependencies")
    if not isinstance(deps, IndexedDict):
        return None
    return _pkg, deps


def _parse_relationships(
    toml_content: IndexedDict[str, ParsedValue],
    packages: list[Package],
) -> list[Relationship]:
    relationships = []
    _pkg: Package | None = None
    toml_pkgs = toml_content.get("package")
    if not isinstance(toml_pkgs, IndexedList):
        return []
    for package in toml_pkgs:
        results = _get_dependencies(package, packages)
        if results:
            _pkg, deps = results
            dependencies = list(deps.keys())

        if _pkg and dependencies:
            for dep in dependencies:
                dep_pkg = next((pkg for pkg in packages if pkg.name == dep), None)
                if dep_pkg:
                    relationships.append(
                        Relationship(
                            from_=dep_pkg,
                            to_=_pkg,
                            type=RelationshipType.DEPENDENCY_OF_RELATIONSHIP,
                        ),
                    )

    return relationships



from pydantic import BaseModel, Field, field_validator


class Advisory(BaseModel):
    cpes: list[str]
    description: str | None = Field(min_length=1)
    epss: float
    id: str = Field(min_length=1)
    namespace: str = Field(min_length=1)
    percentile: float
    severity: str = Field(min_length=1)
    urls: list[str]
    version_constraint: str | None = Field()
    cvss3: str | None = Field()
    cvss4: str | None = Field()
    cwe_ids: list[str] | None = None
    cve_finding: str | None = None
    auto_approve: bool = False

    @field_validator("cpes")
    @classmethod
    def check_cpes_min_length(cls, value: list[str]) -> list[str]:
        for cpe in value:
            if len(cpe) < 1:
                error_message = "Each cpe string must be at least 1 character long."
                raise ValueError(error_message)
        return value

    @field_validator("urls")
    @classmethod
    def check_urls_min_length(cls, value: list[str]) -> list[str]:
        for url in value:
            if len(url) < 1:
                error_message = "Each url string must be at least 1 character long."
                raise ValueError(error_message)
        return value

    def get_info_count(self) -> int:
        info_count = 0
        for attr in [
            self.cpes,
            self.description,
            self.urls,
            self.version_constraint,
        ]:
            if attr:
                info_count += 1
        return info_count

    def __repr__(self) -> str:
        return f"Advisory(id={self.id}, namespace={self.namespace}, severity={self.severity})"



import logging
import re
from itertools import (
    zip_longest,
)

LOGGER = logging.getLogger(__name__)

INFINITE = "//INFINITE//"
PLATFORM_SPECIFIERS = {"macos", "darwin", "linux", "win", "mingw"}


def _simplify_pre_release(pre_release: str) -> str | None:
    if any(platform_specifier in pre_release for platform_specifier in PLATFORM_SPECIFIERS):
        return None
    return pre_release


def normalize(version: str) -> tuple[list[str], str | None]:
    # Split the version into parts and handle the pre-release part
    version_split = version.split("-", maxsplit=1) if version.count("-") >= 1 else [version]
    version_parts = list(re.split(r"[\-\.]", version_split[0]))
    pre_release = version_split[1] if len(version_split) > 1 else None
    simplified_pre_release = _simplify_pre_release(pre_release.lower()) if pre_release else None
    return version_parts, simplified_pre_release


def _aux_compare_tokens(token1: str, token2: str) -> bool | None:
    if token1.isdigit() and token2.isdigit():
        return int(token1) > int(token2)
    if token1.isdigit() or token1 == INFINITE:
        return True
    if token2.isdigit() or token2 == INFINITE:
        return False
    # Compare non-numeric qualifiers
    if token1 != token2:
        return token1 > token2
    return None


def _aux_compare_suffix(suffix1: str, suffix2: str) -> bool | None:
    if suffix1 and not suffix2:
        return True
    if suffix2 and not suffix1:
        return False
    if suffix1 != suffix2:
        return suffix1 > suffix2
    return None


def compare_tokens(token1: str, token2: str) -> bool | None:
    if "+" in token1 or "+" in token2:
        t1_parts = token1.split("+", 1)
        t2_parts = token2.split("+", 1)
        num1 = t1_parts[0]
        num2 = t2_parts[0]
        suffix1 = t1_parts[1] if len(t1_parts) > 1 else ""
        suffix2 = t2_parts[1] if len(t2_parts) > 1 else ""
        if num1.isdigit() and num2.isdigit():
            if int(num1) != int(num2):
                return int(num1) > int(num2)
        elif num1 != num2:
            return num1 > num2
        result = _aux_compare_suffix(suffix1, suffix2)
        if result is not None:
            return result

    return _aux_compare_tokens(token1, token2)


def compare_pre_releases(v1_pre_release: str | None, v2_pre_release: str | None) -> bool:
    if v1_pre_release is None and v2_pre_release is not None:
        return True
    if v1_pre_release is not None and v2_pre_release is None:
        return False
    if v1_pre_release and v2_pre_release:
        return v1_pre_release > v2_pre_release
    return False


def compare_versions(*, version1: str, version2: str, include_same: bool = False) -> bool:
    """Compare two version ranges.

    Returns True if version1 is higher or equal than version2.
    """
    version_1, v1_pre_release = normalize(version1)
    version_2, v2_pre_release = normalize(version2)

    if include_same and version_1 == version_2 and v1_pre_release == v2_pre_release:
        return True

    for part1, part2 in zip_longest(version_1, version_2, fillvalue="0"):
        if part1 != part2:
            comparison = compare_tokens(part1, part2)
            if comparison is not None:
                return comparison

    if (
        not v2_pre_release
        and all(specifier == "0" for specifier in version_2)
        and all(specifier == "0" for specifier in version_1)
    ):
        return True

    if include_same:
        return compare_pre_releases(v1_pre_release, v2_pre_release)

    return False


def parse_version_range(version_range: str) -> tuple[str, ...]:
    if version_range.startswith(("<", "<=")):
        version_range = ">0 " + version_range
    if version_range.startswith((">", ">=")) and "<" not in version_range:
        version_range = version_range + f" <{INFINITE}"

    range_pattern = re.compile(r"([<>]=?)\s*([^<>=\s]+)")
    matches = range_pattern.findall(version_range)
    operators_and_values = [item for pair in matches for item in pair]
    return tuple(operators_and_values)


def is_single_version(range_str: str) -> bool:
    operator_pattern = re.compile(r"[<>]")
    return not bool(operator_pattern.search(range_str))


def is_single_version_in_range(version: str, _range: str) -> bool:
    single_version = version.lstrip("=")
    op1, min2, op2, max2 = parse_version_range(_range)
    include_same_1 = "=" in op1
    include_same_2 = "=" in op2
    return compare_versions(
        version1=single_version,
        version2=min2,
        include_same=include_same_1,
    ) and compare_versions(
        version1=max2,
        version2=single_version,
        include_same=include_same_2,
    )


def do_ranges_intersect(range1: str, range2: str) -> bool:
    """Compare two version ranges.

    Returns True if the input ranges intersect
    """
    is_single_version1 = is_single_version(range1)
    is_single_version2 = is_single_version(range2)
    if is_single_version1 and is_single_version2:
        return range1 == range2

    if is_single_version1:
        return is_single_version_in_range(range1, range2)

    if is_single_version2:
        return is_single_version_in_range(range2, range1)

    op1_r1, lower1, op2_r1, upper1 = parse_version_range(range1)
    op1_r2, lower2, op2_r2, upper2 = parse_version_range(range2)
    include_same_1 = bool("=" in op1_r1 and "=" in op2_r2)
    include_same_2 = bool("=" in op1_r2 and "=" in op2_r1)

    return compare_versions(
        version1=upper2,
        version2=lower1,
        include_same=include_same_1,
    ) and compare_versions(
        version1=upper1,
        version2=lower2,
        include_same=include_same_2,
    )


def convert_asterisk_to_range(version: str) -> str:
    semver_pattern = re.compile(r"^(\d+)(\.(\*|\d+)(\.(\*|\d+))?)?$")
    match = semver_pattern.match(version)
    if match:
        major, _, minor, _, patch = match.groups()
        if patch == "*":
            return f">={major}.{minor}.0 <{major}.{minor}.{INFINITE}"
        if minor == "*":
            return f">={major}.0.0 <{major}.{INFINITE}.0"
    return version


def increment_version(version: str, position: int) -> str:
    parts = version.split(".")
    if len(parts) < 2:
        parts.extend(["0"] * (2 - len(parts)))
        position -= 1

    if position < len(parts):
        parts[position] = INFINITE
        for idx in range(position + 1, len(parts)):
            parts[idx] = "0"
    return ".".join(parts)


def simplify_final_version(version: str) -> str:
    if "+" in version:
        version = version.rsplit("+", 1)[0]
    ranges = version.lower().split(".")
    if "final" in ranges[-1]:
        return ".".join(ranges[:-1])
    return version


def convert_semver_to_range(version: str) -> str:
    version = version.replace("==", "=")
    version = version.replace(" ", "")
    if version.startswith("~"):
        version = version[1:]
        return f">={version} <{increment_version(version, 2)}"

    if version.startswith("^"):
        version = version[1:]
        return f">={version} <{increment_version(version, 1)}"

    if ".*" in version:
        return convert_asterisk_to_range(version)

    if is_single_version(version) and not version.startswith("="):
        return f"={simplify_final_version(version)}"

    return simplify_final_version(version)


def convert_to_range(version: str) -> list[str]:
    and_ranges = re.split(r"[,\s]+", version)
    return [convert_semver_to_range(and_range.strip()) for and_range in and_ranges]


def match_version_ranges(dep_version: str, vulnerable_version: str) -> bool:
    and_ranges = convert_to_range(dep_version)
    vulnerable_ranges = vulnerable_version.split("||")
    for vuln_range in vulnerable_ranges:
        vuln_range = convert_semver_to_range(vuln_range.strip())  # noqa: PLW2901
        if all(do_ranges_intersect(and_range, vuln_range) for and_range in and_ranges):
            return True
    return False


def match_vulnerable_versions(dep_version: str, advisory_range: str | None) -> bool:
    if advisory_range is None:
        return False
    pattern = r"([<>]=?|!=)\s+"
    dep_version = re.sub(pattern, r"\1", dep_version)
    try:
        dep_versions = dep_version.replace("||", "|").split("|")
        if any(
            match_version_ranges(dep_version.strip(), advisory_range)
            for dep_version in dep_versions
        ):
            return True
    except ValueError:
        err_msg = (
            f"Error in simple match. Dependency version: {dep_version}, "
            f"Advisory vulnerable versions: {advisory_range}, "
        )
        LOGGER.exception(err_msg)
        return False
    else:
        return False



from urllib.parse import parse_qs

from labels.model.release import Release

PURL_QUALIFIER_DISTRO = "distro"


def purl_qualifiers(
    qualifiers: dict[str, str | None],
    release: Release | None = None,
) -> dict[str, str]:
    if release:
        distro_qualifiers = []
        if release.id_:
            distro_qualifiers.append(release.id_)
        if release.version_id:
            distro_qualifiers.append(release.version_id)
        elif release.build_id:
            distro_qualifiers.append(release.build_id)

        if distro_qualifiers:
            qualifiers[PURL_QUALIFIER_DISTRO] = "-".join(distro_qualifiers)

    return {
        key: qualifiers.get(key, "") or ""
        for key in sorted(qualifiers.keys())
        if qualifiers.get(key)
    }


def extract_distro_info(pkg_str: str) -> tuple[str | None, str | None, str | None]:
    parts = pkg_str.split("?", 1)
    if len(parts) != 2:
        return None, None, None

    query = parts[1]
    params = parse_qs(query)
    distro_id = params.get("distro_id", [None])[0]
    distro_version = params.get("distro_version_id", [None])[0]
    arch = params.get("arch", [None])[0]
    return distro_id, distro_version, arch



import logging
import re
from concurrent.futures import ThreadPoolExecutor
from multiprocessing import cpu_count
from typing import TextIO, cast

from pydantic import BaseModel, ConfigDict

from labels.model.file import LocationReadCloser
from labels.model.package import Package
from labels.model.relationship import Relationship, RelationshipType
from labels.model.resolver import Resolver
from labels.parsers.cataloger.debian.model import DpkgDBEntry
from labels.parsers.cataloger.debian.package import new_dpkg_package
from labels.parsers.cataloger.debian.parse_dpkg_info_files import parse_dpkg_conffile_info
from labels.parsers.cataloger.generic.parser import Environment
from labels.utils.file import parse_bytes

LOGGER = logging.getLogger(__name__)


class DpkgField(BaseModel):
    name: str
    value: str
    model_config = ConfigDict(frozen=True)


def _split_deps(value: str) -> list[str]:
    fields = value.split(",")
    return [field.strip() for field in fields if field.strip()]


def extract_source_version(source: str) -> tuple[str, str | None]:
    if match_result := re.compile(r"(?P<name>\S+)( \((?P<version>.*)\))?").match(source):
        result = match_result.groupdict()
        return result["name"], result["version"] or ""
    return source, ""


def handle_new_key_value(line: str) -> tuple[str | None, str]:
    if ":" in line and not line.startswith(" "):
        key, value = line.split(":", 1)
        value = value.strip()
        match key:
            case "Installed-Size":
                value = str(parse_bytes(value))
        return key, value

    return None, line


def parse_dpkg_status(reader: TextIO) -> list[DpkgDBEntry]:
    content = reader.read()
    entries = []
    for package in content.strip().split("\n\n"):
        lines = package.split("\n")
        data = {}
        prev_key = ""
        for line in lines:
            try:
                key, value = handle_new_key_value(line)
            except ValueError:
                LOGGER.exception("Value error occurred")
                continue
            if key is not None:
                data[key] = value
                prev_key = key
            elif prev_key in data:
                data[prev_key] = f"{data[prev_key]}\n{value}"
        if all(not value for value in data.values()):
            continue
        source_name, source_version = extract_source_version(data.get("Source", ""))
        entry = DpkgDBEntry(
            package=data.get("Package", ""),
            source=source_name,
            version=data.get("Version", ""),
            source_version=source_version or "",
            architecture=data.get("Architecture", ""),
            maintainer=data.get("Maintainer", ""),
            installed_size=int(data.get("Installed-Size", "0")),
            description=data.get("Description", "").strip(),
            provides=_split_deps(data.get("Provides", "")),
            dependencies=_split_deps(data.get("Depends", "")),
            pre_dependencies=_split_deps(data.get("Pre-Depends", "")),
            files=parse_dpkg_conffile_info(data.get("Conffiles", "")),
        )
        entries.append(entry)

    return entries


def parse_dpkg_db(
    resolver: Resolver | None,
    _env: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    entries: list[DpkgDBEntry] = parse_dpkg_status(reader.read_closer)
    with ThreadPoolExecutor(max_workers=cpu_count()) as executor:
        result = [
            pkg if isinstance(pkg, Package) else (pkg[0], pkg[1])
            for pkg in filter(
                None,
                executor.map(
                    lambda entry: new_dpkg_package(
                        entry,
                        reader.location,
                        resolver,
                        _env.linux_release if _env else None,
                    ),
                    entries,
                ),
            )
            if isinstance(pkg, Package | tuple)
        ]
        packages = [item for pkg in result for item in (pkg if isinstance(pkg, tuple) else [pkg])]
    return packages, associate_relationships(packages)


def associate_relationships(
    pkgs: list[Package],
) -> list[Relationship]:
    lookup: dict[str, list[Package]] = {}
    relationships: list[Relationship] = []
    for pkg in pkgs:
        lookup.setdefault(pkg.name, []).append(pkg)
    for pkg in pkgs:
        metadata = cast(DpkgDBEntry, pkg.metadata)
        for provides in metadata.provides or []:
            key = strip_version_specifier(provides)
            lookup.setdefault(key, []).append(pkg)

    for pkg in pkgs:
        metadata = cast(DpkgDBEntry, pkg.metadata)
        all_deps = [
            *(metadata.dependencies or []),
            *(metadata.pre_dependencies or []),
        ]
        for dep_specifier in all_deps:
            deps = split_package_choices(dep_specifier)
            for dep in deps:
                relationships.extend(
                    Relationship(
                        from_=dep_pkg,
                        to_=pkg,
                        type=RelationshipType.DEPENDENCY_OF_RELATIONSHIP,
                    )
                    for dep_pkg in lookup.get(dep, [])
                )

    return relationships


def strip_version_specifier(item: str) -> str:
    # Define the characters that indicate the start of a version specifier
    specifiers = "[(<>="

    # Find the index of the first occurrence of any specifier character
    index = next((i for i, char in enumerate(item) if char in specifiers), None)

    # If no specifier character is found, return the original string
    if index is None:
        return item.strip()

    # Return the substring up to the first specifier character, stripped of
    # leading/trailing whitespace
    return item[:index].strip()


def split_package_choices(value: str) -> list[str]:
    fields = value.split("|")
    return [strip_version_specifier(field) for field in fields if field.strip()]



import logging
import re
from copy import deepcopy
from typing import NotRequired, TypedDict

from pydantic import ValidationError

from labels.model.file import LocationReadCloser
from labels.model.package import Language, Package, PackageType
from labels.model.relationship import Relationship, RelationshipType
from labels.model.resolver import Resolver
from labels.parsers.cataloger.generic.parser import Environment
from labels.parsers.cataloger.javascript.model import YarnLockEntry
from labels.parsers.cataloger.javascript.package import package_url
from labels.utils.strings import format_exception

LOGGER = logging.getLogger(__name__)


def _resolve_pair(line: str) -> tuple[str, str]:
    line = line.strip()
    if ": " in line:
        key, value = line.split(": ")
        return key.strip(), value.strip()

    key, value = line.split(" ", maxsplit=1)
    return key.strip(), value.strip()


def _count_indentation(line: str) -> int:
    # Stripping the leading spaces and comparing the length difference
    return len(line) - len(line.lstrip(" "))


def _is_start_of_list(
    current_package: str | None,
    current_package_version: str | None,
    line: str,
) -> bool:
    return bool(
        current_package and current_package_version and ":" in line and line.strip().endswith(":"),
    )


def _is_list_item(
    current_package: str | None,
    current_package_version: str | None,
    current_key: str | None,
    current_indentation: int | None,
    line: str,
) -> bool:
    return bool(
        current_package
        and current_package_version
        and current_key
        and current_indentation
        and _count_indentation(line) > current_indentation,
    )


class YarnPackage(TypedDict):
    checksum: str
    dependencies: NotRequired[list[tuple[str, str]]]
    integrity: NotRequired[str]
    line: int
    resolution: NotRequired[str]
    resolved: NotRequired[str]
    version: str


def parse_current_package(line: str, index: int) -> tuple[str | None, int | None]:
    line = line.strip()
    if match_ := re.match(r'^"?((?:@\w[\w\-\.]*/)?\w[\w\-\.]*)@', line):
        current_package = match_.groups()[0]
        current_package_line = index
    else:
        current_package = None
        current_package_line = None

    return current_package, current_package_line


def _parse_yarn_file(yarn_lock_content: str) -> dict[tuple[str, str], YarnPackage]:
    yarn_lock_lines = yarn_lock_content.strip().split("\n")

    # Dictionary to store the parsed yarn lock data
    parsed_yarn_lock = {}

    # Temporary variables for current package and dependencies
    current_package: str | None = None
    current_package_version: str | None = None
    current_indentation = None
    current_key = None

    # Iterate through each line and parse the content
    for index, line in enumerate(yarn_lock_lines, 1):
        if not line:
            current_indentation = None
            continue
        if line.startswith("#"):
            continue
        if not line.startswith(" "):
            current_package, current_package_line = parse_current_package(line, index)
            current_package_version = None
        if current_package and line.strip().startswith("version"):
            _, raw_version = _resolve_pair(line)
            current_package_version = raw_version.strip('"')
            parsed_yarn_lock[(current_package, current_package_version)] = {
                "line": current_package_line,
                "version": current_package_version,
            }
        elif _is_start_of_list(current_package, current_package_version, line):
            current_indentation = _count_indentation(line)
            current_key = line.strip().split(":")[0]
            parsed_yarn_lock[(current_package, current_package_version)][  # type: ignore
                current_key
            ] = []  # type: ignore
        elif _is_list_item(
            current_package,
            current_package_version,
            current_key,
            current_indentation,
            line,
        ):
            parsed_yarn_lock[(current_package, current_package_version)][  # type: ignore
                current_key  # type: ignore
            ].append(_resolve_pair(line))
        elif current_package and current_package_version:
            current_indentation = None
            key, value = _resolve_pair(line)
            parsed_yarn_lock[(current_package, current_package_version)][key] = value.strip('"')

    return parsed_yarn_lock  # type: ignore


def _get_name(pkg_info: tuple[str, str], item: YarnPackage) -> str:
    if resolution := item.get("resolution"):
        is_scoped_package = resolution.startswith("@")
        if is_scoped_package:
            return f"@{resolution.split('@')[1]}"
        return resolution.split("@")[0]

    return pkg_info[0]


def _extract_packages(
    parsed_yarn_lock: dict[tuple[str, str], YarnPackage],
    reader: LocationReadCloser,
) -> list[Package]:
    packages = []
    for pkg_info, item in parsed_yarn_lock.items():
        name = _get_name(pkg_info, item)
        version = item.get("version")

        if not name or not version:
            continue

        current_location = deepcopy(reader.location)
        if current_location.coordinates:
            current_location.coordinates.line = item["line"]
        try:
            packages.append(
                Package(
                    name=name,
                    version=version,
                    locations=[current_location],
                    language=Language.JAVASCRIPT,
                    licenses=[],
                    type=PackageType.NpmPkg,
                    p_url=package_url(name, version),
                    metadata=YarnLockEntry(
                        resolved=item.get("resolved"),
                        integrity=item.get("integrity"),
                    ),
                ),
            )
        except ValidationError as ex:
            LOGGER.warning(
                "Malformed package. Required fields are missing or data types are incorrect.",
                extra={
                    "extra": {
                        "exception": format_exception(str(ex)),
                        "location": current_location.path(),
                    },
                },
            )
            continue

    return packages


def _extract_relationships(
    parsed_yarn_lock: dict[tuple[str, str], YarnPackage],
    packages: list[Package],
) -> list[Relationship]:
    relationships = []
    for pkg_info, item in parsed_yarn_lock.items():
        current_pkg = next(
            (package for package in packages if package.name == _get_name(pkg_info, item)),
            None,
        )

        if current_pkg is None:
            continue

        if "dependencies" in item:
            for raw_dep_name, _ in item["dependencies"]:
                dep_name = raw_dep_name.strip('"')
                # TO-DO: check if the version matches
                if dep := next(
                    (package for package in packages if package.name == dep_name),
                    None,
                ):
                    relationships.append(
                        Relationship(
                            from_=dep,
                            to_=current_pkg,
                            type=RelationshipType.DEPENDENCY_OF_RELATIONSHIP,
                        ),
                    )
    return relationships


def parse_yarn_lock(
    _resolver: Resolver | None,
    _environment: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    parsed_yarn_lock = _parse_yarn_file(reader.read_closer.read())
    packages = _extract_packages(parsed_yarn_lock, reader)
    relationships = _extract_relationships(parsed_yarn_lock, packages)
    return packages, relationships



import logging
from typing import TYPE_CHECKING

from labels.model.file import Scope
from labels.utils.strings import format_exception

if TYPE_CHECKING:  # pragma: no cover
    from collections.abc import ItemsView

from typing import cast

from pydantic import ValidationError

from labels.model.file import Location, LocationReadCloser
from labels.model.indexables import IndexedDict, ParsedValue
from labels.model.package import Language, Package, PackageType
from labels.model.relationship import Relationship
from labels.model.resolver import Resolver
from labels.parsers.cataloger.generic.parser import Environment
from labels.parsers.cataloger.python.package import package_url
from labels.parsers.collection.json import parse_json_with_tree_sitter

LOGGER = logging.getLogger(__name__)


def _get_location(location: Location, sourceline: int, *, is_dev: bool) -> Location:
    location.scope = Scope.DEV if is_dev else Scope.PROD
    if location.coordinates:
        c_upd = {"line": sourceline}
        l_upd = {"coordinates": location.coordinates.model_copy(update=c_upd)}
        return location.model_copy(update=l_upd)
    return location


def _get_version(value: IndexedDict[str, ParsedValue]) -> str:
    version = value.get("version")
    if not isinstance(version, str):
        return ""
    return version.strip("=<>~^ ")


def _get_packages(
    reader: LocationReadCloser,
    dependencies: ParsedValue | None,
    *,
    is_dev: bool = False,
) -> list[Package]:
    if dependencies is None or not isinstance(dependencies, IndexedDict):
        return []

    packages = []

    items: ItemsView[str, ParsedValue] = dependencies.items()
    for name, value in items:
        if not isinstance(value, IndexedDict) or not isinstance(name, str):
            continue
        version = _get_version(value)
        if not name or not version:
            continue

        location = _get_location(reader.location, value.position.start.line, is_dev=is_dev)

        try:
            packages.append(
                Package(
                    name=name,
                    version=version,
                    locations=[location],
                    language=Language.PYTHON,
                    type=PackageType.PythonPkg,
                    p_url=package_url(
                        name=name,
                        version=version,
                        package=None,
                    ),
                    licenses=[],
                    is_dev=is_dev,
                ),
            )
        except ValidationError as ex:
            LOGGER.warning(
                "Malformed package. Required fields are missing or data types are incorrect.",
                extra={
                    "extra": {
                        "exception": format_exception(str(ex)),
                        "location": location.path(),
                    },
                },
            )
            continue

    return packages


def parse_pipfile_lock_deps(
    _resolver: Resolver | None,
    _env: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    content = cast(
        "IndexedDict[str, ParsedValue]",
        parse_json_with_tree_sitter(reader.read_closer.read()),
    )
    deps: ParsedValue | None = content.get("default")
    dev_deps: ParsedValue | None = content.get("develop")
    packages = [
        *_get_packages(reader, deps),
        *_get_packages(reader, dev_deps, is_dev=True),
    ]
    return packages, []



import logging

from packageurl import PackageURL
from pydantic import ValidationError

from labels.model.file import Location
from labels.model.package import Language, Package, PackageType
from labels.parsers.cataloger.python.model import PythonPackage
from labels.parsers.cataloger.python.parse_wheel_egg_metadata import ParsedData
from labels.utils.strings import format_exception

LOGGER = logging.getLogger(__name__)


def new_package_for_package(
    data: ParsedData,
    sources: Location,
) -> Package | None:
    name = data.python_package.name
    version = data.python_package.version

    if not name or not version:
        return None

    try:
        return Package(
            name=name,
            version=version,
            p_url=package_url(
                name,
                version,
                data.python_package,
            ),
            locations=[sources],
            language=Language.PYTHON,
            type=PackageType.PythonPkg,
            metadata=data.python_package,
            licenses=[],
        )
    except ValidationError as ex:
        LOGGER.warning(
            "Malformed package. Required fields are missing or data types are incorrect.",
            extra={
                "extra": {
                    "exception": format_exception(str(ex)),
                    "location": sources.path(),
                },
            },
        )
        return None


def package_url(name: str, version: str, package: PythonPackage | None) -> str:
    return PackageURL(
        type="pypi",
        namespace="",
        name=name,
        version=version,
        qualifiers=_purl_qualifiers_for_package(package),
        subpath="",
    ).to_string()


def _purl_qualifiers_for_package(
    package: PythonPackage | None,
) -> dict[str, str]:
    if not package:
        return {}
    if (
        hasattr(package, "direct_url_origin")
        and package.direct_url_origin
        and package.direct_url_origin.vcs
    ):
        url = package.direct_url_origin
        return {"vcs_url": f"{url.vcs}+{url.url}@{url.commit_id}"}
    return {}



import os
import re
import uuid
from pathlib import Path
from urllib.parse import urlunparse

from labels.model.advisories import Advisory
from labels.model.core import SbomConfig, SourceType
from labels.model.package import HealthMetadata, Language, Package, PackageType
from labels.model.relationship import Relationship, RelationshipType
from labels.model.resolver import Resolver
from labels.resolvers.container_image import ContainerImage
from labels.resolvers.directory import Directory


def is_valid_email(email: str) -> bool:
    email_regex = r"^[a-zA-Z0-9._%+-]+@[a-zA-Z0-9.-]+\.[a-zA-Z]{2,}$"
    return re.match(email_regex, email) is not None


def get_author_info(
    health_metadata: HealthMetadata,
) -> list[tuple[str | None, str | None]]:
    author_info = []

    if health_metadata.authors:
        authors_list = health_metadata.authors.split(", ")
        for author in authors_list:
            name: str | None = None
            email: str | None = None
            email_match = re.search(r"<([^<>]+)>", author)

            if email_match:
                email_candidate: str = email_match.group(1)
                email = email_candidate if is_valid_email(email_candidate) else None
                name = author.replace(email_match.group(0), "").strip() or None
            else:
                name = author.strip() or None

            author_info.append((name, email))

    return author_info


def sanitize_name(text: str) -> str:
    return re.sub(r"[^a-zA-Z0-9.-]", "-", text)


def get_document_namespace(working_dir: str) -> str:
    input_type = "unknown-source-type"

    if Path(working_dir).is_file():
        input_type = "file"
    elif Path(working_dir).is_dir():
        input_type = "dir"

    unique_id = uuid.uuid4()
    identifier = os.path.join(input_type, str(unique_id))
    if working_dir != ".":
        identifier = os.path.join(input_type, f"{working_dir}-{unique_id}")

    return urlunparse(
        ("https", "fluidattacks.com", identifier, "", "", ""),
    )


def get_relative_path(absolute_path: str) -> str:
    root_path = Path(absolute_path)
    current_dir = Path.cwd()
    try:
        return str(root_path.relative_to(current_dir))
    except ValueError:
        return absolute_path


def set_namespace_version(config: SbomConfig, resolver: Resolver) -> tuple[str, str | None]:
    namespace = ""
    version = None
    if config.source_type == SourceType.DIRECTORY and isinstance(resolver, Directory):
        namespace = get_relative_path(resolver.root)
    if config.source_type in {SourceType.DOCKER, SourceType.ECR} and isinstance(
        resolver,
        ContainerImage,
    ):
        namespace = resolver.context.image_ref
        version = resolver.context.id
    return namespace, version


def create_packages_for_test() -> tuple[list[Package], list[Relationship]]:
    test_from_pkg = Package(
        name="source-package",
        version="1.0.0",
        locations=[],
        licenses=[],
        language=Language.PYTHON,
        type=PackageType.PythonPkg,
        p_url="pkg:test/source@1.0.0",
        metadata=None,
        found_by="python-pyproject-toml-cataloger",
        advisories=[
            Advisory(
                id="test-advisory",
                urls=["https://test-url.com"],
                severity="HIGH",
                epss=0.75,
                percentile=8.5,
                description="Test advisory description",
                version_constraint=">=1.0.0",
                cpes=[],
                namespace="github",
                cvss3=None,
                cvss4=None,
            ),
        ],
    )
    test_to_pkg = Package(
        name="target-package",
        version="1.0.0",
        locations=[],
        licenses=[],
        language=Language.PYTHON,
        type=PackageType.PythonPkg,
        p_url="pkg:test/target@1.0.0",
        metadata=None,
        found_by="python-pyproject-toml-cataloger",
    )
    test_packages = [
        Package(
            name="test-package",
            version="1.0.0",
            locations=[],
            licenses=[],
            language=Language.PYTHON,
            type=PackageType.PythonPkg,
            p_url="pkg:python/test-package@1.0.0",
            metadata=None,
            found_by="python-pyproject-toml-cataloger",
        ),
        test_from_pkg,
        test_to_pkg,
    ]
    test_relationships = [
        Relationship(
            from_=test_from_pkg,
            to_=test_to_pkg,
            type=RelationshipType.DEPENDENCY_OF_RELATIONSHIP,
        ),
    ]
    return test_packages, test_relationships



import logging
from copy import deepcopy

from packageurl import PackageURL
from pydantic import ValidationError

from labels.model.file import DependencyType, Location, Scope
from labels.model.indexables import IndexedDict, ParsedValue
from labels.model.package import Language, Package, PackageType
from labels.parsers.cataloger.javascript.model import NpmPackageLockEntry
from labels.utils.strings import format_exception

LOGGER = logging.getLogger(__name__)


def new_package_lock_v1(
    location: Location,
    name: str,
    value: IndexedDict[str, ParsedValue],
    *,
    is_transitive: bool,
) -> Package | None:
    version: str = str(value.get("version", ""))
    if not name or not version:
        return None

    alias_prefix_package_lock = "npm:"
    if version.startswith(alias_prefix_package_lock):
        name, version = version.removeprefix(alias_prefix_package_lock).rsplit(
            "@",
            1,
        )
    current_location = deepcopy(location)
    is_dev = value.get("dev", False)
    current_location.scope = Scope.DEV if is_dev else Scope.PROD
    if current_location.coordinates:
        current_location.coordinates.line = value.position.start.line
        current_location.dependency_type = (
            DependencyType.TRANSITIVE if is_transitive else DependencyType.DIRECT
        )
    try:
        return Package(
            name=name,
            version=version,
            locations=[current_location],
            language=Language.JAVASCRIPT,
            licenses=[],
            type=PackageType.NpmPkg,
            metadata=NpmPackageLockEntry(
                resolved=str(value.get("resolved")) if value.get("resolved") is not None else None,
                integrity=str(value.get("integrity"))
                if value.get("integrity") is not None
                else None,
                is_dev=bool(value.get("dev", False)),
            )
            if value.get("resolved") and "integrity" in value
            else None,
            p_url=package_url(name, version),
        )
    except ValidationError as ex:
        LOGGER.warning(
            "Malformed package. Required fields are missing or data types are incorrect.",
            extra={
                "extra": {
                    "exception": format_exception(str(ex)),
                    "location": current_location.path(),
                },
            },
        )
        return None


def new_package_lock_v2(
    location: Location,
    name: str,
    value: IndexedDict[str, ParsedValue],
    *,
    is_transitive: bool,
) -> Package | None:
    version: str = str(value.get("version", ""))

    if not name or not version:
        return None

    current_location = location
    is_dev = bool(value.get("dev", False))
    current_location.scope = Scope.DEV if is_dev else Scope.PROD
    if current_location.coordinates:
        current_location.coordinates.line = value.position.start.line
        current_location.dependency_type = (
            DependencyType.TRANSITIVE if is_transitive else DependencyType.DIRECT
        )
    try:
        return Package(
            name=name,
            version=version,
            locations=[current_location],
            language=Language.JAVASCRIPT,
            licenses=[],
            type=PackageType.NpmPkg,
            metadata=NpmPackageLockEntry(
                resolved=str(value.get("resolved")) if value.get("resolved") is not None else None,
                integrity=str(value.get("integrity"))
                if value.get("integrity") is not None
                else None,
                is_dev=is_dev,
            ),
            p_url=package_url(name, version),
        )
    except ValidationError as ex:
        LOGGER.warning(
            "Malformed package. Required fields are missing or data types are incorrect.",
            extra={
                "extra": {
                    "exception": format_exception(str(ex)),
                    "location": current_location.path(),
                },
            },
        )
        return None


def package_url(name: str, version: str) -> str:
    namespace = ""
    fields = name.split("/", 2)
    if len(fields) > 1:
        namespace = fields[0]
        name = fields[1]

    if not name:
        return ""

    return PackageURL(  # type: ignore
        type="npm",
        namespace=namespace,
        name=name,
        version=version,
        qualifiers={},
        subpath="",
    ).to_string()



import json
import logging
from datetime import datetime
from enum import Enum
from typing import Any

from pydantic import BaseModel

from labels.model.indexables import IndexedDict, IndexedList
from labels.model.package import Package
from labels.model.relationship import Relationship

LOGGER = logging.getLogger(__name__)


class EnumEncoder(json.JSONEncoder):
    def default(self, item: Any) -> Any:
        if isinstance(item, Enum):
            return item.value
        if isinstance(item, IndexedList):
            return list(item.data)
        if isinstance(item, IndexedDict):
            return dict(item.data)
        if isinstance(item, BaseModel):
            return item.model_dump()
        if isinstance(item, datetime):
            return item.isoformat()
        return super().default(item)


def serialize_package(package: Package) -> dict[str, Any]:
    return {
        "id": package.id_,
        "name": package.name,
        "version": package.version,
        "locations": [
            {
                "path": loc.path(),
                "line": loc.coordinates.line if loc.coordinates and loc.coordinates.line else None,
                "layer": loc.coordinates.file_system_id
                if loc.coordinates and loc.coordinates.file_system_id
                else None,
                "dependency_type": loc.dependency_type.value,
                "scope": loc.scope.value,
                "reachable_cves": loc.reachable_cves,
            }
            for loc in package.locations
        ],
        "licenses": package.licenses,
        "type": package.type.value,
        "language": package.language.value,
        "platform": package.type.get_platform_value(),
        "package_url": package.p_url,
        "found_by": package.found_by,
        "health_metadata": package.health_metadata.model_dump()
        if package.health_metadata
        else None,
        "advisories": [adv.model_dump() for adv in package.advisories or []],
    }


def serialize_packages(packages: list[Package]) -> list[dict[str, Any]]:
    raw_pkgs = [serialize_package(pkg) for pkg in packages]
    return validate_pkgs(raw_pkgs)


def validate_pkgs(raw_sbom_pkgs: list[dict[str, Any]]) -> list[dict[str, Any]]:
    pkgs_to_remove = []
    for index, raw_pkg in enumerate(raw_sbom_pkgs):
        try:
            pkg = json.loads(json.dumps(raw_pkg, cls=EnumEncoder))
            raw_sbom_pkgs[index] = pkg
        except TypeError as err:
            LOGGER.warning(
                "Cannot serialize package at index %d: %s. Error: %s",
                index,
                raw_pkg,
                str(err),
            )
            pkgs_to_remove.append(index)

    for index in sorted(pkgs_to_remove, reverse=True):
        raw_sbom_pkgs.pop(index)
    return raw_sbom_pkgs


def build_relationship_map(relationships: list[Relationship]) -> list[dict[str, str | list[str]]]:
    dependency_map: dict[str, list[str]] = {}
    for rel in relationships:
        from_pkg = rel.from_
        to_pkg = rel.to_
        ref = to_pkg.id_
        dep = from_pkg.id_
        dependency_map.setdefault(ref, []).append(dep)
    return [{"from": ref, "to": deps} for ref, deps in dependency_map.items()]


def build_sbom_metadata(namespace: str, version: str | None, timestamp: str) -> dict[str, Any]:
    return {
        "name": namespace,
        "version": version,
        "timestamp": timestamp,
        "tool": "Fluid-Labels",
        "organization": "Fluid attacks",
    }



import json
import logging
import os
import re
import shutil
import subprocess  # nosec
import tempfile
from pathlib import Path
from typing import Any

from labels.model.sources import ImageContext, ImageMetadata, LayerData
from labels.utils.exceptions import (
    DockerImageNotFoundError,
    InvalidImageReferenceError,
    SkopeoNotFoundError,
)
from labels.utils.file import extract_tar_file

LOGGER = logging.getLogger(__name__)


def _format_image_ref(image_ref: str, *, daemon: bool = False) -> str:
    image_ref_pattern = (
        r"^(?:(?P<host>[\w\.\-]+(?:\:\d+)?)/)?"
        r"(?P<namespace>(?:[\w\.\-]+(?:/[\w\.\-]+)*)?/)?"
        r"(?P<image>[\w\.\-]+)(?::(?P<tag>[\w\.\-]+))?(?:@"
        r"(?P<digest>sha256:[A-Fa-f0-9]{64}))?$"
    )
    prefix_to_use = "docker-daemon:" if daemon else "docker://"
    prefix_used: str | None = None
    prefixes = ["docker://", "docker-daemon:"]
    for prefix in prefixes:
        if image_ref.startswith(prefix):
            image_ref = image_ref.replace(prefix, "", 1)
            prefix_used = prefix
            break

    prefix_to_use = prefix_used or prefix_to_use

    if re.match(image_ref_pattern, image_ref):
        return f"{prefix_to_use}{image_ref}"

    raise InvalidImageReferenceError(image_ref)


def _get_skopeo_path() -> str:
    skopeo_path = shutil.which("skopeo")
    if not skopeo_path:
        raise SkopeoNotFoundError
    return skopeo_path


def _execute_command(command_args: list[str]) -> bool:
    with subprocess.Popen(  # noqa: S603
        command_args,
        shell=False,
        stdout=subprocess.PIPE,
    ) as proc:
        exit_code = proc.wait()
        return exit_code == 0


def _load_manifest(layers_dir: str) -> dict[str, Any]:
    with Path(os.path.join(layers_dir, "manifest.json")).open(encoding="utf-8") as f:
        return json.load(f)


def _load_config(config_digest: str, layers_dir: str) -> dict[str, Any]:
    path = os.path.join(layers_dir, config_digest.replace("sha256:", ""))
    with Path(path).open(encoding="utf-8") as f:
        return json.load(f)


def _extract_layer(layer: dict[str, Any], layers_dir: str, output_dir: str) -> None:
    layer_digest = layer["digest"].replace("sha256:", "")
    tar_path = os.path.join(layers_dir, layer_digest)
    if Path(tar_path).exists():
        dest = os.path.join(output_dir, layer["digest"])
        Path(dest).mkdir(parents=True, exist_ok=True)
        extract_tar_file(tar_path, dest)


def _build_copy_auth_args(
    *,
    unauthenticated_args: list[str],
    username: str | None = None,
    password: str | None = None,
    token: str | None = None,
    aws_creds: str | None = None,
) -> list[str]:
    if username and password:
        unauthenticated_args.extend(
            ["--src-username", username, "--src-password", password],
        )

    elif token:
        unauthenticated_args.extend(["--src-registry-token", token])

    elif aws_creds:
        unauthenticated_args.append(f"--src-creds={aws_creds}")

    return unauthenticated_args


def _build_inspect_auth_args(
    *,
    unauthenticated_args: list[str],
    username: str | None = None,
    password: str | None = None,
    token: str | None = None,
    aws_creds: str | None = None,
) -> list[str]:
    if username and password:
        unauthenticated_args.extend(["--username", username, "--password", password])

    elif token:
        unauthenticated_args.append(f"--registry-token={token}")

    elif aws_creds:
        unauthenticated_args.append(f"--creds={aws_creds}")

    return unauthenticated_args


def _custom_object_hook(json_object: dict[str, Any]) -> ImageMetadata | dict:
    if "Name" in json_object and "Digest" in json_object and "RepoTags" in json_object:
        layersdata = [
            LayerData(
                mimetype=layer_data["MIMEType"],
                digest=layer_data["Digest"],
                size=layer_data["Size"],
                annotations=layer_data.get("Annotations"),
            )
            for layer_data in json_object["LayersData"]
        ]
        return ImageMetadata(
            name=json_object["Name"],
            digest=json_object["Digest"],
            repotags=json_object["RepoTags"],
            created=json_object["Created"],
            dockerversion=json_object["DockerVersion"],
            labels=json_object["Labels"],
            architecture=json_object["Architecture"],
            os=json_object["Os"],
            layers=json_object["Layers"],
            layersdata=layersdata,
            env=json_object["Env"],
        )
    return json_object


def copy_image(  # noqa: PLR0913
    image_ref: str,
    dest_path: str,
    *,
    username: str | None = None,
    password: str | None = None,
    aws_creds: str | None = None,
    token: str | None = None,
) -> bool:
    skopeo_path = _get_skopeo_path()

    formated_image_ref = _format_image_ref(image_ref)

    command_args = [
        skopeo_path,
        "copy",
        "--dest-decompress",
        "--src-tls-verify=false",
        "--insecure-policy",
        "--override-os",
        "linux",
        formated_image_ref,
        f"dir:{dest_path}",
    ]

    authenticated_args = _build_copy_auth_args(
        unauthenticated_args=command_args,
        username=username,
        password=password,
        aws_creds=aws_creds,
        token=token,
    )

    return _execute_command(authenticated_args)


def extract_docker_image(  # noqa: PLR0913
    image: ImageMetadata,
    output_dir: str,
    *,
    username: str | None = None,
    password: str | None = None,
    token: str | None = None,
    aws_creds: str | None = None,
    daemon: bool = False,
) -> tuple[str, dict[str, Any]]:
    layers_dir_temp = tempfile.mkdtemp()

    formated_image_ref = _format_image_ref(image.image_ref, daemon=daemon)

    copy_image(
        image_ref=formated_image_ref or image.image_ref,
        dest_path=layers_dir_temp,
        username=username,
        password=password,
        token=token,
        aws_creds=aws_creds,
    )

    manifest = _load_manifest(layers_dir_temp)
    manifest["config_full"] = _load_config(manifest["config"]["digest"], layers_dir_temp)

    for layer in manifest["layers"]:
        _extract_layer(layer, layers_dir_temp, output_dir)

    return layers_dir_temp, manifest


def get_docker_image(  # noqa: PLR0913
    image_ref: str,
    *,
    username: str | None = None,
    password: str | None = None,
    token: str | None = None,
    aws_creds: str | None = None,
    daemon: bool = False,
) -> ImageMetadata:
    skopeo_path = _get_skopeo_path()

    formated_image_ref = _format_image_ref(image_ref, daemon=daemon)

    command_args = [
        skopeo_path,
        "inspect",
        "--tls-verify=false",
        "--override-os",
        "linux",
        formated_image_ref,
    ]

    authenticated_args = _build_inspect_auth_args(
        unauthenticated_args=command_args,
        username=username,
        password=password,
        aws_creds=aws_creds,
        token=token,
    )

    try:
        result = subprocess.run(  # noqa: S603
            authenticated_args,
            check=True,
            capture_output=True,
            text=True,
        )

        image_metadata: ImageMetadata = json.loads(
            result.stdout,
            object_hook=_custom_object_hook,
        )
        if image_metadata:
            image_metadata = image_metadata.model_copy(
                update={"image_ref": image_ref},
            )

        return image_metadata
    except subprocess.CalledProcessError as error:
        raise DockerImageNotFoundError(image_ref, error.stderr.strip()) from error


def get_image_context(  # noqa: PLR0913
    *,
    image: ImageMetadata,
    username: str | None = None,
    password: str | None = None,
    token: str | None = None,
    aws_creds: str | None = None,
    daemon: bool = False,
) -> ImageContext:
    temp_dir = tempfile.mkdtemp()

    layers_dir, manifest = extract_docker_image(
        image,
        temp_dir,
        username=username,
        password=password,
        token=token,
        aws_creds=aws_creds,
        daemon=daemon,
    )

    return ImageContext(
        id=image.digest,
        name=image.name,
        publisher="",
        arch=image.architecture,
        size=str(sum(x.size for x in image.layersdata)),
        full_extraction_dir=temp_dir,
        layers_dir=layers_dir,
        manifest=manifest,
        image_ref=image.image_ref,
    )



import asyncio

from labels.integrations.docker import get_docker_image, get_image_context
from labels.integrations.ecr import ecr_connection, get_token
from labels.model.core import SbomConfig, SourceType
from labels.model.sources import AwsCredentials, AwsRole, ImageMetadata
from labels.resolvers.container_image import ContainerImage
from labels.resolvers.directory import Directory
from labels.utils.exceptions import (
    AwsCredentialsNotFoundError,
    AwsExternalIdNotDefinedError,
    AwsRoleNotDefinedError,
    UnexpectedSBOMSourceError,
)


def _get_async_ecr_connection(role: AwsRole, source: str) -> tuple[str, ImageMetadata]:
    return asyncio.run(ecr_connection(role, source))


def _handle_directory(sbom_config: SbomConfig) -> Directory:
    return Directory(
        root=sbom_config.source,
        include=sbom_config.include,
        exclude=sbom_config.exclude,
    )


def _handle_docker(*, sbom_config: SbomConfig, is_daemon: bool = False) -> ContainerImage:
    docker_image = get_docker_image(
        sbom_config.source,
        username=sbom_config.docker_user,
        password=sbom_config.docker_password,
        daemon=is_daemon,
    )

    context = get_image_context(
        image=docker_image,
        username=sbom_config.docker_user,
        password=sbom_config.docker_password,
        daemon=is_daemon,
    )

    return ContainerImage(img=docker_image, context=context)


def _handle_ecr(sbom_config: SbomConfig) -> ContainerImage:
    if not sbom_config.aws_role:
        raise AwsRoleNotDefinedError

    if not sbom_config.aws_external_id:
        raise AwsExternalIdNotDefinedError

    role = AwsRole(
        external_id=sbom_config.aws_external_id,
        role=sbom_config.aws_role,
    )

    token, image_metadata = _get_async_ecr_connection(
        role=role,
        source=sbom_config.source,
    )

    context = get_image_context(
        image=image_metadata,
        aws_creds=f"AWS:{token}",
    )

    return ContainerImage(img=image_metadata, context=context)


def handle_ecr_with_credentials(
    image_uri: str,
    aws_credentials: AwsCredentials,
) -> ContainerImage:
    token = asyncio.run(get_token(aws_credentials))
    aws_token = f"AWS:{token}"
    image_metadata = get_docker_image(image_ref=f"docker://{image_uri}", aws_creds=aws_token)
    context = get_image_context(image=image_metadata, aws_creds=aws_token)

    return ContainerImage(img=image_metadata, context=context)


def resolve_sbom_source(sbom_config: SbomConfig) -> Directory | ContainerImage:
    match sbom_config.source_type:
        case SourceType.DIRECTORY:
            return _handle_directory(sbom_config)
        case SourceType.DOCKER:
            return _handle_docker(sbom_config=sbom_config)
        case SourceType.DOCKER_DAEMON:
            return _handle_docker(sbom_config=sbom_config, is_daemon=True)
        case SourceType.ECR:
            return _handle_ecr(sbom_config)
        case SourceType.ECR_WITH_CREDENTIALS:
            if not sbom_config.aws_credentials:
                raise AwsCredentialsNotFoundError

            return handle_ecr_with_credentials(
                image_uri=sbom_config.source,
                aws_credentials=sbom_config.aws_credentials,
            )
        case _:
            raise UnexpectedSBOMSourceError(sbom_config.source_type)



import tarfile
from contextlib import suppress

from labels.model.file import Coordinates, Location

# IEC Sizes
BYTE = 1 << (0 * 10)
KIBYTE = 1 << (1 * 10)
MIBYTE = 1 << (2 * 10)
GIBYTE = 1 << (3 * 10)
TIBYTE = 1 << (4 * 10)
PIBYTE = 1 << (5 * 10)
EIBYTE = 1 << (6 * 10)

# SI Sizes
IBYTE = 1
KBYTE = IBYTE * 1000
MBYTE = KBYTE * 1000
GBYTE = MBYTE * 1000
TBYTE = GBYTE * 1000
PBYTE = TBYTE * 1000
EBYTE = PBYTE * 1000

bytes_size_table = {
    "b": BYTE,
    "kib": KIBYTE,
    "kb": KBYTE,
    "mib": MIBYTE,
    "mb": MBYTE,
    "gib": GIBYTE,
    "gb": GBYTE,
    "tib": TIBYTE,
    "tb": TBYTE,
    "pib": PIBYTE,
    "pb": PBYTE,
    "eib": EIBYTE,
    "eb": EBYTE,
    # Without suffix
    "": BYTE,
    "ki": KIBYTE,
    "k": KBYTE,
    "mi": MIBYTE,
    "m": MBYTE,
    "gi": GIBYTE,
    "g": GBYTE,
    "ti": TIBYTE,
    "t": TBYTE,
    "pi": PIBYTE,
    "p": PBYTE,
    "ei": EIBYTE,
    "e": EBYTE,
}


def parse_bytes(size_human: str) -> int:
    last_digit = 0
    has_comma = False

    for i, char in enumerate(size_human):
        if not (char.isdigit() or char in {".", ","}):
            break
        if char == ",":
            has_comma = True
        last_digit = i + 1

    num_str = size_human[:last_digit]
    if has_comma:
        num_str = num_str.replace(",", "")

    try:
        num = float(num_str)
    except ValueError as exc:
        error_message = f"Could not parse number: {num_str}"
        raise ValueError(error_message) from exc

    extra = size_human[last_digit:].strip().lower()
    if extra in bytes_size_table:
        num *= bytes_size_table[extra]
        return int(num)

    error_message = f"Unhandled size name: {extra}"
    raise ValueError(error_message)


def extract_tar_file(tar_path: str, out_path: str) -> None:
    with tarfile.open(tar_path, "r") as tar:
        for member in tar.getmembers():
            member.name = member.name.lstrip("/")
            with suppress(Exception):
                tar.extract(member, path=out_path)


def new_location_from_image(
    access_path: str | None,
    layer_id: str,
    real_path: str | None = None,
) -> Location:
    if access_path and not access_path.startswith("/"):
        access_path = f"/{access_path}"
    return Location(
        coordinates=Coordinates(real_path=real_path or "", file_system_id=layer_id),
        access_path=access_path,
        annotations={},
    )


def new_location(real_path: str) -> Location:
    return Location(
        coordinates=Coordinates(
            real_path=real_path,
        ),
        access_path=real_path,
        annotations={},
    )



def generate_cpe(package_manager: str, package_name: str, vulnerable_version: str) -> str:
    part = "a"
    vendor = package_name.split(":")[0] if ":" in package_name else "*"
    product = package_name.lower()
    version = vulnerable_version
    language = package_manager
    update = edition = sw_edition = target_sw = target_hw = other = "*"

    return (
        f"cpe:2.3:{part}:{vendor}:{product}:{version}:{update}:{edition}:"
        f"{language}:{sw_edition}:{target_sw}:{target_hw}:{other}"
    )



from collections.abc import Callable

from pydantic import BaseModel, ConfigDict

from labels.model.file import LocationReadCloser
from labels.model.package import Package
from labels.model.relationship import Relationship
from labels.model.release import Release
from labels.model.resolver import Resolver


class Environment(BaseModel):
    linux_release: Release | None
    model_config = ConfigDict(frozen=True)


Parser = Callable[
    [Resolver, Environment, LocationReadCloser],
    tuple[list[Package], list[Relationship]] | None,
]



import logging
from copy import deepcopy
from typing import cast

from labels.model.file import LocationReadCloser
from labels.model.indexables import IndexedDict, ParsedValue
from labels.model.package import Package
from labels.model.relationship import Relationship, RelationshipType
from labels.model.resolver import Resolver
from labels.parsers.cataloger.generic.parser import Environment
from labels.parsers.cataloger.php.model import PhpComposerInstalledEntry, PhpComposerLockEntry
from labels.parsers.cataloger.php.package import new_package_from_composer
from labels.parsers.collection.json import parse_json_with_tree_sitter

LOGGER = logging.getLogger(__name__)
EMPTY_DICT: IndexedDict[str, ParsedValue] = IndexedDict()


def parse_composer_lock(
    _: Resolver | None,
    __: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    package_json: IndexedDict[str, ParsedValue] = cast(
        IndexedDict[str, ParsedValue],
        parse_json_with_tree_sitter(reader.read_closer.read()),
    )
    packages: list[Package] = []
    relationships: list[Relationship] = []
    pkgs: IndexedDict[str, ParsedValue] = cast(
        IndexedDict[str, ParsedValue],
        package_json.get("packages", EMPTY_DICT),
    )
    pkgs_dev: IndexedDict[str, ParsedValue] = cast(
        IndexedDict[str, ParsedValue],
        package_json.get("packages-dev", EMPTY_DICT),
    )
    for is_dev, package in [
        *[(False, x) for x in pkgs],
        *[(True, x) for x in pkgs_dev],
    ]:
        if not isinstance(package, IndexedDict):
            continue
        new_location = deepcopy(reader.location)

        if pkg := new_package_from_composer(package, new_location, is_dev=is_dev):
            packages.append(pkg)

    for parsed_pkg in packages:
        if not isinstance(parsed_pkg.metadata, PhpComposerInstalledEntry | PhpComposerLockEntry):
            continue
        for dep_name in parsed_pkg.metadata.require or []:
            package_dep = next(
                (x for x in packages if x.name == dep_name),
                None,
            )
            if package_dep:
                relationships.append(
                    Relationship(
                        from_=parsed_pkg,
                        to_=package_dep,
                        type=RelationshipType.DEPENDENCY_OF_RELATIONSHIP,
                    ),
                )
    return packages, relationships



from enum import Enum
from io import TextIOWrapper
from typing import TextIO

from pydantic import BaseModel, ConfigDict


class DependencyType(Enum):
    DIRECT = "DIRECT"
    TRANSITIVE = "TRANSITIVE"
    UNKNOWN = "UNKNOWN"


class Scope(Enum):
    DEV = "DEV"
    PROD = "PROD"


class Type(Enum):
    TYPE_REGULAR = "TypeRegular"
    TYPE_HARD_LINK = "TypeHardLink"
    TYPE_SYM_LINK = "TypeSymLink"
    TYPE_CHARACTER_DEVICE = "TypeCharacterDevice"
    TYPE_BLOCK_DEVICE = "TypeBlockDevice"
    TYPE_DIRECTORY = "TypeDirectory"
    TYPE_FIFO = "TypeFIFO"
    TYPE_SOCKET = "TypeSocket"
    TYPE_IRREGULAR = "TypeIrregular"


class Metadata(BaseModel):
    path: str
    link_destination: str
    user_id: int
    group_id: int
    type: Type
    mime_type: str


class Coordinates(BaseModel):
    real_path: str
    file_system_id: str | None = None
    line: int | None = None

    def __str__(self) -> str:
        result = f"RealPath={self.real_path}"
        if self.file_system_id:
            result += f" Layer={self.file_system_id}"
        return f"Location<{result}>"


class LocationMetadata(BaseModel):
    annotations: dict[str, str]

    def merge(self, other: "LocationMetadata") -> "LocationMetadata":
        return LocationMetadata(annotations={**self.annotations, **other.annotations})


class LocationData(BaseModel):
    coordinates: Coordinates
    access_path: str

    def __hash__(self) -> int:
        return hash((self.access_path, self.coordinates.file_system_id))


class Location(BaseModel):
    scope: Scope = Scope.PROD
    coordinates: Coordinates | None = None
    access_path: str | None = None
    annotations: dict[str, str] | None = None
    dependency_type: DependencyType = DependencyType.UNKNOWN
    reachable_cves: list[str] = []

    def path(self) -> str:
        path = self.access_path or (self.coordinates.real_path if self.coordinates else None)
        if not path:
            error_msg = "Both access_path and coordinates.real_path are empty"
            raise ValueError(error_msg)
        return path.strip().replace(" ", "_")


class LocationReadCloser(BaseModel):
    location: Location
    read_closer: TextIO | TextIOWrapper
    model_config = ConfigDict(arbitrary_types_allowed=True)



import logging
from typing import Any

import phpserialize
from pydantic import ValidationError

from labels.model.file import LocationReadCloser
from labels.model.package import Language, Package, PackageType
from labels.model.relationship import Relationship
from labels.model.resolver import Resolver
from labels.parsers.cataloger.generic.parser import Environment
from labels.parsers.cataloger.php.package import package_url_from_pecl
from labels.utils.strings import format_exception

LOGGER = logging.getLogger(__name__)


def php_to_python(obj: Any) -> Any:
    if isinstance(obj, phpserialize.phpobject):
        return {k: php_to_python(v) for k, v in obj.__php_vars__.items()}
    if isinstance(obj, dict):
        return {php_to_python(k): php_to_python(v) for k, v in obj.items()}
    if isinstance(obj, list):
        return [php_to_python(i) for i in obj]

    return obj


def parse_pecl_serialized(
    _: Resolver | None,
    __: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    packages: list[Package] = []
    relationships: list[Relationship] = []

    unserialized_data = phpserialize.loads(reader.read_closer.read().encode(), decode_strings=True)
    parsed_data = php_to_python(unserialized_data)
    name = parsed_data.get("name")
    version = parsed_data.get("version", {}).get("release")

    if not name or not version:
        return [], []

    try:
        packages.append(
            Package(
                name=name,
                version=version,
                locations=[reader.location],
                language=Language.PHP,
                licenses=[],
                type=PackageType.PhpPeclPkg,
                metadata=None,
                p_url=package_url_from_pecl(name, version),
            ),
        )
        return packages, relationships
    except ValidationError as ex:
        LOGGER.warning(
            "Malformed package. Required fields are missing or data types are incorrect.",
            extra={
                "extra": {
                    "exception": format_exception(str(ex)),
                    "location": reader.location.path(),
                },
            },
        )
        return [], []



import logging
import re
from copy import deepcopy

from pydantic import BaseModel, ValidationError

from labels.model.file import Location, LocationReadCloser
from labels.model.package import Language, Package, PackageType
from labels.model.relationship import Relationship
from labels.model.resolver import Resolver
from labels.parsers.cataloger.generic.parser import Environment
from labels.parsers.cataloger.java.model import JavaArchive, JavaPomProject
from labels.parsers.cataloger.java.package import package_url
from labels.utils.strings import format_exception

LOGGER = logging.getLogger(__name__)
# Constants
QUOTE = r'["\']'
NL = r"(\n?\s*)?"
TEXT = r"[a-zA-Z0-9._-]+"
RE_GRADLE_KTS: re.Pattern[str] = re.compile(
    rf"(runtimeOnly|api|compile|plugin|compileOnly|implementation)\({QUOTE}"
    rf"(?P<group>{TEXT}):(?P<name>{TEXT}):"
    rf"(?P<version>{TEXT})"
    rf"(?::(?P<classifier>{TEXT}))?"
    rf"{QUOTE}\)",
)


def is_comment(line: str) -> bool:
    return (
        line.strip().startswith("//")
        or line.strip().startswith("/*")
        or line.strip().endswith("*/")
    )


class LockFileDependency(BaseModel):
    group: str
    name: str
    version: str
    line: int | None = None


def parse_gradle_lockfile_kts(
    _resolver: Resolver | None,
    __: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    dependencies = parse_dependencies(reader)
    packages = create_packages(dependencies, reader.location)
    return packages, []


def parse_dependencies(reader: LocationReadCloser) -> list[LockFileDependency]:
    dependencies: list[LockFileDependency] = []
    is_block_comment = False

    for line_no, raw_line in enumerate(reader.read_closer.readlines(), start=1):
        line = raw_line.strip()
        is_block_comment = update_block_comment_status(line, is_block_comment=is_block_comment)

        if is_block_comment or is_comment(line):
            continue

        dependency = extract_dependency(line, line_no)
        if dependency:
            dependencies.append(dependency)

    return dependencies


def update_block_comment_status(line: str, *, is_block_comment: bool) -> bool:
    if "/*" in line:
        return True
    if "*/" in line:
        return False
    return is_block_comment


def extract_dependency(line: str, line_no: int) -> LockFileDependency | None:
    if match := RE_GRADLE_KTS.match(line):
        version = match.group("version")
        if version:
            return LockFileDependency(
                group=match.group("group"),
                name=match.group("name"),
                version=version,
                line=line_no,
            )

    return None


def create_packages(
    dependencies: list[LockFileDependency],
    reader_location: Location,
) -> list[Package]:
    packages: list[Package] = []

    for dependency in dependencies:
        name = dependency.name
        version = dependency.version

        if not name or not version:
            continue

        location = deepcopy(reader_location)
        if location.coordinates:
            location.coordinates.line = dependency.line

        archive = create_java_archive(dependency, name, version)

        package = create_package(name, version, location, archive, dependency.group)
        if package:
            packages.append(package)

    return packages


def create_java_archive(
    dependency: LockFileDependency,
    name: str,
    version: str,
) -> JavaArchive:
    return JavaArchive(
        pom_project=JavaPomProject(
            group_id=dependency.group,
            name=name,
            artifact_id=name,
            version=version,
        ),
    )


def create_package(
    name: str,
    version: str,
    location: Location,
    archive: JavaArchive,
    group: str,
) -> Package | None:
    try:
        return Package(
            name=f"{group}:{name}",
            version=version,
            locations=[location],
            language=Language.JAVA,
            type=PackageType.JavaPkg,
            metadata=archive,
            p_url=package_url(name, version, archive),
            licenses=[],
        )
    except ValidationError as ex:
        LOGGER.warning(
            "Malformed package. Required fields are missing or data types are incorrect.",
            extra={
                "extra": {
                    "exception": format_exception(str(ex)),
                    "location": location.path(),
                },
            },
        )
        return None



import re
from collections.abc import Iterator
from copy import deepcopy
from typing import cast

from packageurl import PackageURL
from pydantic import BaseModel

from labels.model.file import DependencyType, Location, LocationReadCloser, Scope
from labels.model.package import Language, Package, PackageType
from labels.model.relationship import Relationship
from labels.model.resolver import Resolver
from labels.parsers.cataloger.generic.parser import Environment


class GolangModuleEntry(BaseModel):
    h1_digest: str


GO_DIRECTIVE: re.Pattern[str] = re.compile(
    r"(?P<directive>require|replace) \(",
)
GO_MOD_DEP: re.Pattern[str] = re.compile(
    r"^\s+(?P<product>.+?/[\w\-\.~]+?)\sv(?P<version>\S+)",
)
GO_REPLACE: re.Pattern[str] = re.compile(
    r"^\s+(?P<old_prod>.+?/[\w\-\.~]+?)(\sv(?P<old_ver>\S+))?\s=>"
    r"\s(?P<new_prod>.+?/[\w\-\.~]+?)(\sv(?P<new_ver>\S+))?$",
)
GO_REP_DEP: re.Pattern[str] = re.compile(
    r"replace\s(?P<old_prod>.+?/[\w\-\.~]+?)(\sv(?P<old_ver>\S+))?\s=>"
    r"\s(?P<new_prod>.+?/[\w\-\.~]+?)(\sv(?P<new_ver>\S+))?$",
)
GO_REQ_MOD_DEP: re.Pattern[str] = re.compile(
    r"require\s(?P<product>.+?/[\w\-\.~]+?)\sv(?P<version>\S+)",
)
GO_VERSION: re.Pattern[str] = re.compile(
    r"\ngo (?P<major>\d)\.(?P<minor>\d+)(\.\d+)?\n",
)


def parse_go_sum_file(
    resolver: Resolver | None,
    reader: LocationReadCloser,
) -> dict[str, str] | None:
    out = {}

    if resolver is None or not reader.location.coordinates:
        return None

    go_sum_path = reader.location.coordinates.real_path.replace(".mod", ".sum")
    go_sum_location = resolver.relative_file_path(reader.location, go_sum_path)

    if go_sum_location is None:
        return None

    try:
        contents = resolver.file_contents_by_location(go_sum_location)
    except OSError:
        return None
    if not contents:
        return None
    # Read the contents of the file
    lines = contents.read().splitlines()

    for line in lines:
        parts = line.split(" ")
        if len(parts) < 3:
            continue

        name_version = f"{parts[0]} {parts[1]}"
        hash_value = parts[2]

        out[name_version] = hash_value

    return out


def package_url(module_name: str, module_version: str) -> str:
    fields = module_name.split("/")

    if all(not f.strip() for f in fields):
        return ""

    namespace = ""
    name = ""
    subpath = ""

    if len(fields) == 1:
        name = fields[0]
    elif len(fields) == 2:
        name = fields[1]
        namespace = fields[0]
    else:
        name = fields[2]
        namespace = "/".join(fields[:2])
        subpath = "/".join(fields[3:])

    return PackageURL(  # type: ignore
        type="golang",
        namespace=namespace,
        name=name,
        version=module_version,
        qualifiers=None,
        subpath=subpath,
    ).to_string()


def add_require(
    matched: re.Match[str],
    req_dict: dict[str, Package],
    line_number: int,
    parent_location: Location,
    sums: dict[str, str] | None,
) -> None:
    product: str = matched.group("product")
    version: str = matched.group("version")
    location = deepcopy(parent_location)
    location.scope = Scope.PROD
    if location.coordinates:
        location.coordinates.line = line_number
        location.dependency_type = DependencyType.DIRECT
    sum_key = f"{product} v{version}"
    metadata = GolangModuleEntry(h1_digest=sums[sum_key]) if sums and sum_key in sums else None
    req_dict[product] = Package(
        name=product,
        version=version,
        type=PackageType.GoModulePkg,
        locations=[location],
        p_url=package_url(product, version),
        metadata=metadata,
        language=Language.GO,
        licenses=[],
        is_dev=False,
    )


def replace_req(
    req_dict: dict[str, Package],
    replace_list: list[tuple[re.Match[str], int]],
    parent_location: Location,
    sums: dict[str, str] | None,
) -> Iterator[Package]:
    for matched, line_number in replace_list:
        match_dict = cast(dict[str, str], matched.groupdict())
        old_pkg, old_version = match_dict["old_prod"], match_dict["old_ver"]
        repl_pkg, version = match_dict["new_prod"], match_dict["new_ver"]

        if old_pkg not in req_dict:
            continue

        if old_version and not version:
            version = req_dict[old_pkg].version

        if not version or (old_version and req_dict[old_pkg].version != old_version):
            continue

        location = deepcopy(parent_location)

        location.scope = Scope.PROD
        if location.coordinates:
            location.coordinates.line = line_number
            location.dependency_type = DependencyType.DIRECT

        sum_key = f"{repl_pkg} v{version}"
        metadata = GolangModuleEntry(h1_digest=sums[sum_key]) if sums and sum_key in sums else None

        req_dict[old_pkg] = Package(
            name=repl_pkg,
            version=version,
            type=PackageType.GoModulePkg,
            locations=[location],
            p_url=package_url(repl_pkg, version),
            metadata=metadata,
            language=Language.GO,
            licenses=[],
            is_dev=False,
        )

    return iter(req_dict.values())


def resolve_go_deps(
    content: str,
    location: Location,
    sums: dict[str, str] | None,
) -> Iterator[Package]:
    go_req_directive: str = ""
    replace_list: list[tuple[re.Match[str], int]] = []
    req_dict: dict[str, Package] = {}

    for line_number, line in enumerate(content.splitlines(), 1):
        if matched := GO_REQ_MOD_DEP.search(line):
            add_require(matched, req_dict, line_number, location, sums)
        elif replace := GO_REP_DEP.search(line):
            replace_list.append((replace, line_number))
        elif not go_req_directive:
            if directive := GO_DIRECTIVE.match(line):
                go_req_directive = directive.group("directive")
        elif go_req_directive == "replace":
            if replace := GO_REPLACE.search(line):
                replace_list.append((replace, line_number))
                continue
            go_req_directive = ""
        elif matched := GO_MOD_DEP.search(line):
            add_require(matched, req_dict, line_number, location, sums)
        else:
            go_req_directive = ""
    return replace_req(req_dict, replace_list, location, sums)


def parse_go_mod(
    resolver: Resolver | None,
    __: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    packages = []
    sums = parse_go_sum_file(resolver, reader)
    content = reader.read_closer.read()
    go_version = GO_VERSION.search(content)
    if not go_version:
        return [], []
    major = int(cast(dict[str, str], go_version.groupdict())["major"])
    minor = int(cast(dict[str, str], go_version.groupdict())["minor"])
    if major >= 2 or (major == 1 and minor >= 17):
        packages = list(resolve_go_deps(content, reader.location, sums))
    return packages, []



import logging
import os
import stat
import tempfile
import zipfile
from pathlib import Path

from pydantic import ValidationError

from labels.model.file import LocationReadCloser
from labels.model.package import Language, Package, PackageType
from labels.model.relationship import Relationship
from labels.model.resolver import Resolver
from labels.parsers.cataloger.generic.parser import Environment
from labels.parsers.cataloger.java.model import JavaArchive, JavaPomProperties
from labels.parsers.cataloger.java.package import package_url
from labels.utils.strings import format_exception

LOGGER = logging.getLogger(__name__)


def is_safe_path(base_path: str, target_path: str) -> bool:
    base_path = os.path.normpath(base_path)
    target_path = os.path.normpath(target_path)
    return os.path.commonpath([base_path]) == os.path.commonpath([base_path, target_path])


def safe_extract(apk_file: zipfile.ZipFile, destination: str) -> None:
    for file_info in apk_file.infolist():
        file_name = file_info.filename
        if Path(file_name).is_absolute() or file_name.startswith(("..", "./")):
            continue

        target_path = os.path.join(destination, file_name)

        if not is_safe_path(destination, target_path):
            continue

        if (file_info.external_attr >> 16) & stat.S_IFLNK:
            continue

        try:
            apk_file.extract(file_name, destination)
        except Exception:
            LOGGER.exception("Error extracting %s", file_name)


def parse_apk(
    _resolver: Resolver | None,
    _env: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    packages: list[Package] = []
    with tempfile.TemporaryDirectory() as output_folder:
        try:
            with zipfile.ZipFile(reader.read_closer.name, "r") as apk_file:
                safe_extract(apk_file, output_folder)
        except zipfile.BadZipFile:
            return packages, []
        files: list[str] = []
        meta_dir = os.path.join(output_folder, "META-INF")
        if Path(meta_dir).exists():
            files = [
                os.path.join(meta_dir, file)
                for file in os.listdir(meta_dir)
                if file.endswith(".version")
            ]
        for file in files:
            file_path = Path(file)
            with file_path.open(encoding="utf-8") as version_reader:
                version = version_reader.read().strip()
            parts = file_path.name.replace(".version", "").split("_", 1)
            group_id = parts[0]
            artifact_id = parts[1]

            if any(not value for value in (artifact_id, version, group_id)):
                continue

            java_archive = JavaArchive(
                pom_properties=JavaPomProperties(
                    group_id=group_id,
                    artifact_id=artifact_id,
                    version=version,
                ),
            )

            try:
                packages.append(
                    Package(
                        name=artifact_id,
                        version=version,
                        licenses=[],
                        locations=[reader.location],
                        language=Language.JAVA,
                        type=PackageType.JavaPkg,
                        metadata=java_archive,
                        p_url=package_url(artifact_id, version, java_archive),
                    ),
                )
            except ValidationError as ex:
                LOGGER.warning(
                    "Malformed package. Required fields are missing or data "
                    "types are incorrect.",
                    extra={
                        "extra": {
                            "exception": format_exception(str(ex)),
                            "location": reader.location.path(),
                        },
                    },
                )
                continue
    return packages, []



from pydantic import BaseModel, Field

from labels.model.package import Digest, Package, PackageType

JENKINS_PLUGIN_POM_PROPERTIES_GROUP_IDS = [
    "io.jenkins.plugins",
    "org.jenkins.plugins",
    "org.jenkins-ci.plugins",
    "io.jenkins-ci.plugins",
    "com.cloudbees.jenkins.plugins",
]


class PomContext(BaseModel):
    parent_info: dict[str, str] | None
    parent_version_properties: dict[str, str] | None
    manage_deps: dict[str, str] | None


class JavaPomParent(BaseModel):
    group_id: str
    artifact_id: str
    version: str


class JavaPomProject(BaseModel):
    path: str | None = None
    group_id: str | None = None
    artifact_id: str | None = None
    version: str | None = None
    name: str | None = None
    parent: JavaPomParent | None = None
    description: str | None = None
    url: str | None = None


class JavaPomProperties(BaseModel):
    name: str | None = None
    group_id: str | None = None
    artifact_id: str | None = None
    version: str | None = None
    path: str | None = None
    scope: str | None = None
    extra: dict[str, str] = Field(default_factory=dict)

    def pkg_type_indicated(self) -> PackageType:
        if any(
            self.group_id and self.group_id.startswith(prefix)
            for prefix in JENKINS_PLUGIN_POM_PROPERTIES_GROUP_IDS
        ) or (self.group_id and ".jenkins.plugin" in self.group_id):
            return PackageType.JenkinsPluginPkg
        return PackageType.JavaPkg


class JavaManifest(BaseModel):
    main: dict[str, str]
    sections: list[dict[str, str]] | None = None


class JavaArchive(BaseModel):
    virtual_path: str | None = None
    manifest: JavaManifest | None = None
    pom_properties: JavaPomProperties | None = None
    pom_project: JavaPomProject | None = None
    archive_digests: list[Digest] = Field(default_factory=list)
    parent: Package | None = None
